# -*- coding: utf-8 -*-.
"""
Created on Tue Jan 21 13:04:58 2020

@author: xavier.mouy
"""
from ecosound.core.annotation import Annotation
import ecosound.core.tools
import pandas as pd
import xarray as xr
import os

class Measurement(Annotation):

    def __init__(self, measurer_name=None, measurer_version=None, measurements_name=None):
        """ Measurement object.

        Object to "store" sound measurements. Inheritate all methods from the
        ecosound Annotaion class.

        Parameters
        ----------
        measurer_name : str, optional
            Name of the measurer that was used to calculate the measurements.
            The default is None. 
        measurer_version : str, optional
            Version of the measurer that was used to calculate the measurements.
            The default is None.
        measurements_name : list of str, optional
            List with the name of each measurement. The default is None.

        Returns
        -------
        None. ecosound Measurement object with a .data and .metadata dataframes

        """
        super(Measurement, self).__init__()
        metadata = {'measurer_name': measurer_name,
                    'measurer_version': measurer_version,
                    'measurements_name': [measurements_name],
                    }
        self._metadata = pd.DataFrame(metadata)
        self.data = pd.concat([self.data,pd.DataFrame(columns=metadata['measurements_name'][0])])

    @property
    def metadata(self):
        """
        Return the metadata attribute.
        
        Includes adictionary with the measurer_name, measurer_version, and
        measurements_name.
        """
        return self._metadata

    def to_netcdf(self, file):
        """
        Write measurement data to a netcdf file.

        Write measurementss as .nc file. This format works well with xarray
        and Dask.

        Parameters
        ----------
        file : str
            Path of the netcdf file (.nc) to be written.

        Returns
        -------
        None.
        """

        if file.endswith('.nc') == False:
            file = file + '.nc'
        self._enforce_dtypes()
        meas = self.data
        meas.set_index('time_min_date', drop=False, inplace=True)
        meas.index.name = 'date'
        dxr1 = meas.to_xarray()
        dxr1.attrs['datatype'] = 'Measurement'
        dxr1.attrs['measurements_name'] = self.metadata.measurements_name.values[0]
        dxr1.attrs['measurer_name'] = self.metadata.measurer_name.values[0]
        dxr1.attrs['measurer_version'] = self.metadata.measurer_version.values[0]
        dxr1.to_netcdf(file, engine='netcdf4', format='NETCDF4')

    def from_netcdf(self, file, verbose=False):
        """
        Import measurement data from a netcdf file.

        Load measurements from a .nc file. This format works well with xarray
        and Dask. Only "Measurement" netcdf files created with to_netcdf can
        be imported. all other netcdf files will return an error.

        Parameters
        ----------
        file : str
            Path of the nc file to import. Can be a str if importing a single
            file or entire folder. Needs to be a list if importing multiple 
            files. If 'files' is a folder, all files in that folder ending with
            '.nc' will be imported.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if type(file) is str:
            if os.path.isdir(file):
                file = ecosound.core.tools.list_files(file,
                                                       '.nc',
                                                       recursive=False,
                                                       case_sensitive=True,
                                                       )
                if verbose:
                    print(len(file), 'files found.')
            else:
                file = [file]
        self.data, self._metadata = self._import_netcdf_files(file)
        self.check_integrity(verbose=verbose)    

    def _import_netcdf_files(self, files):
        """Import one or several netcdf files to a Panda datafrane."""
        assert type(files) in (str, list), "Input must be of type str (single \
            file or directory) or list (multiple files)"
        # Import all files to a dataframe
        tmp =[]
        for idx, file in enumerate(files):
            dxr = xr.open_dataset(file)
            if dxr.attrs['datatype'] == 'Measurement':
                if idx == 0:
                    measurer_name = dxr.measurer_name
                    measurer_version = dxr.measurer_version 
                    measurements_name = dxr.measurements_name
                ## check measurere name and version
                if (dxr.measurer_name == measurer_name) & (dxr.measurer_version == measurer_version):
                    tmp2 = dxr.to_dataframe()
                    tmp2.reset_index(inplace=True)
                else:
                    raise ValueError(file + "Not all files were not generated from the same measurer type and version.")
            else:
                raise ValueError(file + 'Not a Measurement file.')
            tmp.append(tmp2)
        data = pd.concat(tmp, ignore_index=True, sort=False)
        data.reset_index(inplace=True, drop=True)
        metadata = {'measurer_name': measurer_name,
                    'measurer_version': measurer_version,
                    'measurements_name': [measurements_name],
                    }
        metadata = pd.DataFrame(metadata)
        return data, metadata
    
    def __add__(self, other):
        """Concatenate data from several Measurement objects."""
        assert type(other) is ecosound.core.measurement.Measurement, "Object type not\
            supported. Can only concatenate Measurement objects together."
        assert other.metadata['measurer_name'].values[0] == self.metadata['measurer_name'].values[0], "Can't concatenate measurements made from different measurers."
        assert other.metadata['measurer_version'].values[0] == self.metadata['measurer_version'].values[0], "Can't concatenate measurements made from different versions of measurers."
        self._enforce_dtypes()
        other._enforce_dtypes()
        self.data = pd.concat([self.data, other.data],
                              ignore_index=True,
                              sort=False)
        return self


# -*- coding: utf-8 -*-
"""
Created on Tue Sep  1 21:34:43 2020

Load NetCDF files for an entire deployment and plot timeseries

@author: xavier.mouy
"""
import xarray


def xr_sort(ds):
    return ds.sortby(ds.time_min_date,ascending=True)

#indir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_in_April_July2019_1342218252'
indir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\test_dataset'

# load data
ds = xarray.open_mfdataset(indir + "/*.nc" ,parallel=True, preprocess=xr_sort)

# sort by ascending date
#ds = ds.sortby(ds.time_min_date,ascending=True)
# filter by class and confidence
#ds_filtered = ds.where(ds.label_class == 'FS').where(ds.confidence>0.7).compute()
#ds_filtered = ds.where((ds.label_class == 'FS') & (ds.confidence>0.7), drop=True)

# Hourly count
#fs_count = ds_filtered.groupby('date.hour').count(dim='date') 
#fs_count2 = ds_filtered.resample(date='H').count(dim='date') 
fs_hourly_count = ds.where((ds.label_class == 'FS') & (ds.confidence>0.7), drop=True).resample(date='H').count(dim='date')
fs_hourly_count.compute()
# plot
fs_hourly_count.label_class.plot()

## min confidence per day
#ds_conf_daily = ds_filtered.confidence.resample(date='D').min(dim='date')
#ds_conf_daily.plot()

#s=[ds_filtered['label_class'].isel(date=slice(0,-1))]

#ds.sortby(time_min_date,ascending=True)

#ds_label = ds.label_class
#ds_conf = ds.confidence

#ds_conf.sortby(date, ascending=True)
#ds_conf_daily = ds_conf.resample(time='D').mean(dim='date')

#ds_conf.mean().compute()

#ds_label.set_index(data["Time"],inplace=True)
# f=ds.groupby('label_class').groups

# ds.sel(date="2019-04-11")


# -*- coding: utf-8 -*-
"""
Created on Thu Oct 15 17:38:07 2020

@author: xavier.mouy
"""
import xarray
import numpy as np
import os
import time
import datetime as dt

def xr_sort(ds):
    #return ds.sortby(ds.time_min_date,ascending=True)
    ds = ds.sortby(ds.date,ascending=True)
    
    return ds.sortby(ds.date,ascending=True)

def to_drop_vars(ds):
    return [x for x in list(ds.variables) if x not in ['date','audio_file_start_date','time_min_date','label_class','label_subclass','confidence']] 

def fix_duplicates(ds):
    index=range(0,len(ds['date']))
    _, unique_index = np.unique(ds['date'], return_index=True)
    index = np.delete(index,unique_index)
    if len(index)>0:
        rnd = np.round(np.random.rand(len(index))*100)-50
        rnd[rnd==0]=1
        rnd_offset =[np.timedelta64(int(x),'ns') for x in rnd]
        time_ax = ds['date'].data
        time_ax[np.array(index)] = time_ax[index] + rnd_offset
        ds['date'] = time_ax
    #ds.time_min_date = time_ax
    return ds

def shift_max_dates(ds,files_dur_sec):
    time_ax = ds['date'].data    
    max_date = ds['audio_file_start_date'][0].data + np.timedelta64(files_dur_sec,'s')
    index = time_ax >= max_date
    time_ax[np.array(index)] = max_date - np.timedelta64(1,'ns')    
    ds['date'] = time_ax    
    return ds

def preprocess_func(ds):
    files_dur_sec = 1799
    return shift_max_dates(xr_sort(fix_duplicates(ds.drop(to_drop_vars(ds)))),files_dur_sec) 


#_, index = np.unique(f['time'], return_index=True)

indir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_Out_Jan_April2019_67391492'
#indir=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\test_dataset'
outfile ='hourly_summary.nc'
#confidence_step = 0.02
confidence_step = 0.24

# Start timer
tic = time.perf_counter()

# load data
print('')
print('Loading dataset')
tic1 = time.perf_counter()
ds = xarray.open_mfdataset(indir + "/*.nc" ,
                            parallel=True,
                            preprocess=preprocess_func,
                            data_vars=['time_min_date','audio_file_start_date','label_class','label_subclass','confidence'],
                            coords=['date'],
                            #combine='nested',
                            #join='outer',
                            )
toc1 = time.perf_counter()
print(f"Executed in {toc1 - tic1:0.4f} seconds")


#ds = xarray.open_dataset(r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\Out_Jan_April2019.nc')

# Extract hourly summaries for each class and threshold
print('')
print('Calculating hourly detections')
tic2 = time.perf_counter()
class_groups = ds.groupby('label_class')
confidence_thresholds = np.arange(1/len(class_groups),1,confidence_step)
class_names = []
Xarrays = []
for name, group in class_groups:
    print('  * Extracting class: ' + name)
    class_names.append(name)
    max_threshold_reached = False
    for conf_idx, threshold in enumerate(confidence_thresholds):
        if max_threshold_reached == False: # 
            detec_thresholded=group.where((group.confidence>=threshold),drop=False)
            if len(detec_thresholded['date'])>0: # if there are detection for that threshold
                detec = detec_thresholded.resample(date='H').count(dim='date')
                print('    - confidence: ', '%.2f'%threshold)
                if conf_idx == 0:
                    # define date array
                    date_array = detec['date']
                    # initialize detec array with nan
                    detec_array = np.empty([len(confidence_thresholds),detec.dims['date']])
                    detec_array[:] = np.NaN            
                # fill in appropriate row of detec array with detection counts
                detec_array[conf_idx,:] = detec['label_class'].values
            else: # if there are no detections anymore
                # fill in appropriate row of detec array with zeros
                detec_array[conf_idx,:] = 0
                max_threshold_reached = True
                print('    - confidence: ', '%.2f'%threshold, ' (empty)')
        else: # fills with zeros once max confidence threshold is reached
            # fill in appropriate row of detec array with zeros
            detec_array[conf_idx,:] = 0
            print('    - confidence: ', '%.2f'%threshold, ' (empty)')
    #create xArray for that class
    Xarrays.append(xarray.DataArray(detec_array, coords={'date': date_array, 'confidence': confidence_thresholds}, dims=['confidence','date']))
#Create XArrayDataset
dataset = xarray.Dataset(dict(zip(class_names,Xarrays)))
# Save dataset as netcdf
outdir = os.path.join(indir,'summary')
os.mkdir(outdir)
dataset.to_netcdf(os.path.join(outdir,outfile))
toc2 = time.perf_counter()
print(f"Executed in {toc2 - tic2:0.4f} seconds")

# Stop timer
toc = time.perf_counter()
print('')
print(f" Overall run time is {toc - tic:0.4f} seconds")


# -*- coding: utf-8 -*-
"""
Created on Fri May 22 14:04:01 2020

@author: xavier.mouy
"""


import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.measurements.measurer_builder import MeasurerFactory
import time

## Input paraneters ##########################################################

audio_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\data\67674121.181017060806.wav"
detection_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\results\67674121.181017060806.wav.nc"

#audio_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\audio_data\AMAR173.4.20190916T011248Z.wav"
#annotation_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\manual_annotations\AMAR173.4.20190916T011248Z.Table.1.selections.txt"
#detection_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset\AMAR173.4.20190916T011248Z.wav.nc"


# Spectrogram parameters
frame = 3000
nfft = 4096
step = 500
#ovlp = 2500
fmin = 0
fmax = 1000
window_type = 'hann'

# start and stop time of wavfile to analyze
t1 = 00#24
t2 = 60#40
## ###########################################################################
tic = time.perf_counter()

# load audio data
sound = Sound(audio_file)
sound.read(channel=0, chunk=[t1, t2], unit='sec', detrend=True)

# Calculates  spectrogram
spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='samp')
spectro.compute(sound, dB=True, use_dask=True, dask_chunks=40)
spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)

# load annotations
annot = Annotation()
annot.from_raven(annotation_file)

# load detections
detec = Measurement()
detec.from_netcdf(detection_file)

toc = time.perf_counter()
print(f"Executed in {toc - tic:0.4f} seconds")

# Plot
graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000, time_min=30)
graph.add_data(sound)
graph.add_annotation(detec, panel=0, color='green', label='Detections')
graph.add_data(spectro)
graph.add_annotation(annot, panel=1, color='red', label='Annotations')
graph.add_annotation(detec, panel=1, color='green', label='Detections')
graph.colormap = 'binary'
#graph.colormap = 'jet'
graph.show()


# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:27:39 2020

@author: xavier.mouy
"""

from .detector_builder import BaseClass
from ecosound.core.annotation import Annotation
from datetime import datetime
import uuid

class Detector1(BaseClass):

    detector_parameters = ('kernel_duration','kernel_bandwidth', 'threshold','duration_min','bandwidth_min')
    
    def __init__(self, *args, **kwargs):
        # Initialize all detector parameters to None
        self.__dict__.update(dict(zip(self.detector_parameters,
                                      [None]*len(self.detector_parameters))))
        # Unpack kwargs as detector parameters if provided on instantiation
        self.__dict__.update(**kwargs)
    
    @property
    def name(self):
        """Return name of the detector."""
        detector_name = 'DetectorTemplate'
        return detector_name
    
    @property
    def version(self):
        """Return version of the detector."""
        version = '0.1'
        return version
    
    def _prerun_check(self, spectrogram):
        # check that all required arguments are defined
        if True in [self.__dict__.get(keys) is None for keys in self.detector_parameters]:
            raise ValueError('Not all detector parameters have been defined.'
                             + ' Required parameters: '
                             + str(self.detector_parameters))
        # check that spectrogram is a spectrogram class
        if not isinstance(spectrogram, Spectrogram):
            raise ValueError('Input must be an ecosound Spectrogram object'
                             + '(ecosound.core.spectrogram).')
    
    def run(self, spectro, debug=False):
        
        self._prerun_check(self, spectrogram)
        
        detec = Annotation()
        detec.data['from_detector'] = True
        detec.data['software_name'] = self.name
        detec.data['software_version'] = self.version
        detec.data['entry_date'] = datetime.now()
        detec.data['uuid'] = detec.data.apply(lambda _: str(uuid.uuid4()), axis=1)
        return detec


# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 22:35:31 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from ecosound.core.spectrogram import Spectrogram
from ecosound.visualization.grapher_builder import GrapherFactory
import time

## Input paraneters ##########################################################

single_channel_file = r"../ecosound/resources/AMAR173.4.20190916T061248Z.wav"

# Spectrogram parameters
frame = 3000
nfft = 4096
step = 500
#ovlp = 2500
fmin = 0 
fmax = 1000
window_type = 'hann'

# start and stop time of wavfile to analyze
t1 = 1515
t2 = 1541
## ###########################################################################
tic = time.perf_counter()

# load audio data
sound = Sound(single_channel_file)
fs = sound.waveform_sampling_frequency
sound.read(channel=0, chunk=[round(t1*fs), round(t2*fs)])

# Calculates  spectrogram
Spectro = Spectrogram(frame, window_type, nfft, step, fs, unit='samp')
Spectro.compute(sound, dB=True, dask=True)

toc = time.perf_counter()
print(f"Executed in {toc - tic:0.4f} seconds")

# # Plot
# graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000)
# graph.add_data(sound)
# graph.add_data(Spectro)
# graph.colormap = 'jet'
# graph.show()


# -*- coding: utf-8 -*-.
"""
Created on Tue Jan 21 13:04:58 2020

@author: xavier.mouy
"""

import pandas as pd
import xarray as xr
import os
import uuid
import warnings
import ecosound.core.tools
import ecosound.core.decorators
from ecosound.core.metadata import DeploymentInfo
import copy


class Annotation():
    """
    A class used for manipulating annotation data.

    The Annotation object stores from both manual analysis annotations
    collected with software like PAMlab and Raven, and outputs from
    automated detectors and classifiers.

    Attributes
    ----------
    data : pandas DataFrame
        Annotation DataFranme.

    Methods
    -------
    check_integrity(verbose=False, time_duplicates_only=False)
        Check integrity of Annotation object.
    from_raven(files, class_header='Sound type',subclass_header=None,
               verbose=False)
        Import annotation data from 1 or several Raven files.
    to_raven(outdir, single_file=False)
        Write annotation data to one or several Raven files.
    from_pamlab(files, verbose=False)
        Import annotation data from 1 or several PAMlab files.
    to_pamlab(outdir, single_file=False)
        Write annotation data to one or several Raven files.
    from_parquet(file)
        Import annotation data from a Parquet file.
    to_parquet(file)
        Write annotation data to a Parquet file.
    from_netcdf(file)
        Import annotation data from a netCDF4 file.
    to_netcdf(file)
        Write annotation data to a netCDF4 file.
    insert_values(**kwargs)
        Manually insert values for given Annotation fields.
    insert_metadata(deployment_info_file)
        Insert metadata information to the annotation from a
        deployment_info_file.
    filter_overlap_with(annot, freq_ovp=True, dur_factor_max=None,
                        dur_factor_min=None,ovlp_ratio_min=None,
                        remove_duplicates=False,inherit_metadata=False,
                        filter_deploymentID=True, inplace=False)
        Filter annotations overalaping with another set of annotations.
    get_labels_class()
        Return all unique class labels.
    get_labels_subclass()
        Return all unique subclass labels.
    get_fields()
        Return list with all annotations fields.
    summary(rows='deployment_ID',columns='label_class')
        Produce a summary pivot table with the number of annotations for two
        given annotation fields.
    __add__()
        Concatenate data from annotation objects uisng the + sign.
    __len__()
        Return number of annotations.
    """

    def __init__(self):
        """
        Initialize Annotation object.

        Sets all the annotation fields.:
            -'uuid': UUID,
                Unique identifier code
            -'from_detector': bool,
                True if data comes from an automatic process.
            -'software_name': str,
                Software name. Can be Raven or PAMlab for manual analysis.
            -'software_version': str,
                Version of the software used to create the annotations.
            -'operator_name': str,
                Name of the person responsible for the creation of the
                annotations.
            -'UTC_offset': float,
                Offset hours to UTC.
            -'entry_date': datetime,
                Date when the annotation was created.
            -'audio_channel': int,
                Channel number.
            -'audio_file_name': str,
                Name of the audio file.
            -'audio_file_dir': str,
                Directory where the audio file is.
            -'audio_file_extension': str,
                Extension of teh audio file.
            -'audio_file_start_date': datetime,
                Date of the audio file start time.
            -'audio_sampling_frequency': int,
                Sampling frequecy of the audio data.
            -'audio_bit_depth': int,
                Bit depth of the audio data.
            -'mooring_platform_name': str,
                Name of the moorig platform (e.g. 'glider','Base plate').
            -'recorder_type': str,
                Name of the recorder type (e.g., 'AMAR'), 'SoundTrap'.
            -'recorder_SN': str,
                Serial number of the recorder.
            -'hydrophone_model': str,
                Model of the hydrophone.
            -'hydrophone_SN': str,
                Serial number of the hydrophone.
            -'hydrophone_depth': float,
                Depth of the hydrophone in meters.
            -'location_name': str,
                Name of the deploymnet location.
            -'location_lat': float,
                latitude of the deployment location in decimal degrees.
            -'location_lon': float,
                longitude of the deployment location in decimal degrees.
            -'location_water_depth': float,
                Water depth at the deployment location in meters.
            -'deployment_ID': str,
                Unique ID of the deployment.
            -'frequency_min': float,
                Minimum frequency of the annotaion in Hz.
            -'frequency_max': float,
                Maximum frequency of the annotaion in Hz.
            -'time_min_offset': float,
                Start time of the annotaion, in seconds relative to the
                begining of the audio file.
            -'time_max_offset': float,
                Stop time of the annotaion, in seconds relative to the
                begining of the audio file.
            -'time_min_date': datetime,
                Date of the annotation start time.
            -'time_max_date': datetime,
                Date of the annotation stop time.
            -'duration': float,
                Duration of the annotation in seconds.
            -'label_class': str,
                label of the annotation class (e.g. 'fish').
            -'label_subclass': str,
                label of the annotation subclass (e.g. 'grunt')
            'confidence': float,
                Confidence of the classification.

        Returns
        -------
        Annotation object.

        """
        self.data = pd.DataFrame({
            'uuid': [],
            'from_detector': [],  # True, False
            'software_name': [],
            'software_version': [],
            'operator_name': [],
            'UTC_offset': [],
            'entry_date': [],
            'audio_channel': [],
            'audio_file_name': [],
            'audio_file_dir': [],
            'audio_file_extension': [],
            'audio_file_start_date': [],
            'audio_sampling_frequency': [],
            'audio_bit_depth': [],
            'mooring_platform_name': [],
            'recorder_type': [],
            'recorder_SN': [],
            'hydrophone_model': [],
            'hydrophone_SN': [],
            'hydrophone_depth': [],
            'location_name': [],
            'location_lat': [],
            'location_lon': [],
            'location_water_depth': [],
            'deployment_ID': [],
            'frequency_min': [],
            'frequency_max': [],
            'time_min_offset': [],
            'time_max_offset': [],
            'time_min_date': [],
            'time_max_date': [],
            'duration': [],
            'label_class': [],
            'label_subclass': [],
            'confidence': []
            })
        self._enforce_dtypes()

    def check_integrity(self, verbose=False, ignore_frequency_duplicates=False):
        """
        Check integrity of Annotation object.

        Tasks performed:
            1- Check that start time < stop time
            2- Check that min frequency < max frequency
            3- Remove duplicate entries based on time and frequency, filename,
               labels and filenames

        Parameters
        ----------
        verbose : bool, optional
            Print summary of the duplicate entries deleted.
            The default is False.
        ignore_frequency_duplicates : bool, optional
            If set to True, doesn't consider frequency values when deleting
            duplicates. It is useful when data are imported from Raven.
            The default is False.

        Raises
        ------
        ValueError
            If annotations have a start time > stop time
            If annotations have a min frequency > max frequency

        Returns
        -------
        None.

        """
        # Drop all duplicates
        count_start = len(self.data)
        if ignore_frequency_duplicates:  # doesn't use frequency boundaries
            self.data = self.data.drop_duplicates(
                subset=['time_min_offset',
                        'time_max_offset',
                        'label_class',
                        'label_subclass',
                        'audio_file_name',
                        ], keep="first",).reset_index(drop=True)
        else:  # remove annot with exact same time AND frequency boundaries
            self.data = self.data.drop_duplicates(
                subset=['time_min_offset',
                        'time_max_offset',
                        'frequency_min',
                        'frequency_max',
                        'label_class',
                        'label_subclass',
                        'audio_file_name',
                        ], keep="first",).reset_index(drop=True)
        count_stop = len(self.data)
        if verbose:
            print('Duplicate entries removed:', str(count_start-count_stop))
        # Check that start and stop times are coherent (i.e. t2 > t1)
        time_check = self.data.index[
            self.data['time_max_offset'] <
            self.data['time_min_offset']].tolist()
        if len(time_check) > 0:
            raise ValueError(
                'Incoherent annotation times (time_min > time_max). \
                 Problematic annotations:' + str(time_check))
        # Check that min and max frequencies are coherent (i.e. fmin < fmax)
        freq_check = self.data.index[
            self.data['frequency_max'] < self.data['frequency_min']].tolist()
        if len(freq_check) > 0:
            raise ValueError(
                'Incoherent annotation frequencies (frequency_min > \
                frequency_max). Problematic annotations:' + str(freq_check))
        if verbose:
            print('Integrity test succesfull')

    def from_raven(self, files, class_header='Sound type', subclass_header=None, verbose=False):
        """
        Import data from 1 or several Raven files.

        Load annotation tables from .txt files generated by the software Raven.

        Parameters
        ----------
        files : str, list
            Path of the txt file(s) to import. Can be a str if importing a single
            file. Needs to be a list if importing multiple files. If 'files' is
            a folder, all files in that folder ending with '.selections.txt'
            will be imported.
        class_header : str, optional
            Name of the header in the Raven file corresponding to the class
            name. The default is 'Sound type'.
        subclass_header : str, optional
            Name of the header in the Raven file corresponding to the subclass
            name. The default is None.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if os.path.isdir(files):
            files = ecosound.core.tools.list_files(files,
                                                   '.selections.txt',
                                                   recursive=False,
                                                   case_sensitive=True,
                                                   )
            if verbose:
                print(len(files), 'annotation files found.')
        data = Annotation._import_csv_files(files)
        files_timestamp = ecosound.core.tools.filename_to_datetime(
            data['Begin Path'].tolist())
        self.data['audio_file_start_date'] = files_timestamp
        self.data['audio_channel'] = data['Channel']
        self.data['audio_file_name'] = data['Begin Path'].apply(
            lambda x: os.path.splitext(os.path.basename(x))[0])
        self.data['audio_file_dir'] = data['Begin Path'].apply(
            lambda x: os.path.dirname(x))
        self.data['audio_file_extension'] = data['Begin Path'].apply(
            lambda x: os.path.splitext(x)[1])
        self.data['time_min_offset'] = data['Begin Time (s)']
        self.data['time_max_offset'] = data['End Time (s)']
        self.data['time_min_date'] = pd.to_datetime(
            self.data['audio_file_start_date'] + pd.to_timedelta(
                self.data['time_min_offset'], unit='s'))
        self.data['time_max_date'] = pd.to_datetime(
            self.data['audio_file_start_date'] +
            pd.to_timedelta(self.data['time_max_offset'], unit='s'))
        self.data['frequency_min'] = data['Low Freq (Hz)']
        self.data['frequency_max'] = data['High Freq (Hz)']
        if class_header is not None:
            self.data['label_class'] = data[class_header]
        if subclass_header is not None:
            self.data['label_subclass'] = data[subclass_header]
        self.data['from_detector'] = False
        self.data['software_name'] = 'raven'
        self.data['uuid'] = self.data.apply(lambda _: str(uuid.uuid4()), axis=1)
        self.data['duration'] = self.data['time_max_offset'] - self.data['time_min_offset']
        self.check_integrity(verbose=verbose, ignore_frequency_duplicates=True)
        if verbose:
                print(len(self), 'annotations imported.')

    def to_raven(self, outdir, outfile='Raven.Table.1.selections.txt', single_file=False):
        """
        Write data to 1 or several Raven files.

        Write annotations as .txt files readable by the software Raven. Output
        files can be written in a single txt file or in several txt files (one
        per audio recording). In the latter case, output file names are
        automatically generated based on the audio file's name.

        Parameters
        ----------
        outdir : str
            Path of the output directory where the Raven files are written.
        outfile : str
            Name of the output file. Only used is single_file is True. The
            default is 'Raven.Table.1.selections.txt'.
        single_file : bool, optional
            If set to True, writes a single output file with all annotations.
            The default is False.

        Returns
        -------
        None.

        """
        if single_file:
            annots = [self.data]
        else:
            annots = [pd.DataFrame(y) for x, y in self.data.groupby(
                'audio_file_name', as_index=False)]
        for annot in annots:
            annot.reset_index(inplace=True, drop=True)
            cols = ['Selection', 'View', 'Channel', 'Begin Time (s)',
                    'End Time (s)', 'Delta Time (s)', 'Low Freq (Hz)',
                    'High Freq (Hz)', 'Begin Path', 'File Offset (s)',
                    'Begin File', 'Class', 'Sound type', 'Software',
                    'Confidence']
            outdf = pd.DataFrame({'Selection': 0, 'View': 0, 'Channel': 0,
                                  'Begin Time (s)': 0, 'End Time (s)': 0,
                                  'Delta Time (s)': 0, 'Low Freq (Hz)': 0,
                                  'High Freq (Hz)': 0, 'Begin Path': 0,
                                  'File Offset (s)': 0, 'Begin File': 0,
                                  'Class': 0, 'Sound type': 0, 'Software': 0,
                                  'Confidence': 0},
                                 index=list(range(annot.shape[0])))
            outdf['Selection'] = range(1, annot.shape[0]+1)
            outdf['View'] = 'Spectrogram 1'
            outdf['Channel'] = annot['audio_channel']
            outdf['Begin Time (s)'] = annot['time_min_offset']
            outdf['End Time (s)'] = annot['time_max_offset']
            outdf['Delta Time (s)'] = annot['duration']
            outdf['Low Freq (Hz)'] = annot['frequency_min']
            outdf['High Freq (Hz)'] = annot['frequency_max']
            outdf['File Offset (s)'] = annot['time_min_offset']
            outdf['Class'] = annot['label_class']
            outdf['Sound type'] = annot['label_subclass']
            outdf['Software'] = annot['software_name']
            outdf['Confidence'] = annot['confidence']
            outdf['Begin Path'] = [os.path.join(x, y) + z for x, y, z
                                   in zip(annot['audio_file_dir'],
                                          annot['audio_file_name'],
                                          annot['audio_file_extension'])]
            outdf['Begin File'] = [x + y for x, y
                                   in zip(annot['audio_file_name'],
                                          annot['audio_file_extension'])]
            outdf = outdf.fillna(0)
            if single_file:
                outfilename = os.path.join(outdir, outfile)
            else:
                outfilename = os.path.join(
                    outdir, str(annot['audio_file_name'].iloc[0]))
                + str(annot['audio_file_extension'].iloc[0])
                + '.chan' + str(annot['audio_channel'].iloc[0])
                + '.Table.1.selections.txt'
            outdf.to_csv(outfilename,
                         sep='\t',
                         encoding='utf-8',
                         header=True,
                         columns=cols,
                         index=False)

    def from_pamlab(self, files, verbose=False):
        """
        Import data from 1 or several PAMlab files.

        Load annotation data from .log files generated by the software PAMlab.

        Parameters
        ----------
        files : str, list
            Path of the txt file to import. Can be a str if importing a single
            file or entire folder. Needs to be a list if importing multiple
            files. If 'files' is a folder, all files in that folder ending with
            'annotations.log' will be imported.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if type(files) is str:
            if os.path.isdir(files):
                files = ecosound.core.tools.list_files(files,
                                                       ' annotations.log',
                                                       recursive=False,
                                                       case_sensitive=True,
                                                       )
                if verbose:
                    print(len(files), 'annotation files found.')
        data = Annotation._import_csv_files(files)
        files_timestamp = ecosound.core.tools.filename_to_datetime(
            data['Soundfile'].tolist())
        self.data['audio_file_start_date'] = files_timestamp
        self.data['operator_name'] = data['Operator']
        self.data['entry_date'] = pd.to_datetime(
            data['Annotation date and time (local)'],
            format='%Y-%m-%d %H:%M:%S.%f')
        self.data['audio_channel'] = data['Channel']
        self.data['audio_file_name'] = data['Soundfile'].apply(
            lambda x: os.path.splitext(os.path.basename(x))[0])
        self.data['audio_file_dir'] = data['Soundfile'].apply(
            lambda x: os.path.dirname(x))
        self.data['audio_file_extension'] = data['Soundfile'].apply(
            lambda x: os.path.splitext(x)[1])
        self.data['audio_sampling_frequency'] = data['Sampling freq (Hz)']
        self.data['recorder_type'] = data['Recorder type']
        self.data['recorder_SN'] = data['Recorder ID']
        self.data['hydrophone_depth'] = data['Recorder depth']
        self.data['location_name'] = data['Station']
        self.data['location_lat'] = data['Latitude (deg)']
        self.data['location_lon'] = data['Longitude (deg)']
        self.data['time_min_offset'] = data['Left time (sec)']
        self.data['time_max_offset'] = data['Right time (sec)']
        self.data['time_min_date'] = pd.to_datetime(
            self.data['audio_file_start_date']
            + pd.to_timedelta(self.data['time_min_offset'], unit='s'))
        self.data['time_max_date'] = pd.to_datetime(
            self.data['audio_file_start_date'] +
            pd.to_timedelta(self.data['time_max_offset'], unit='s'))
        self.data['frequency_min'] = data['Bottom freq (Hz)']
        self.data['frequency_max'] = data['Top freq (Hz)']
        self.data['label_class'] = data['Species']
        self.data['label_subclass'] = data['Call type']
        self.data['from_detector'] = False
        self.data['software_name'] = 'pamlab'
        self.data['uuid'] = self.data.apply(lambda _: str(uuid.uuid4()), axis=1)
        self.data['duration'] = self.data['time_max_offset'] - self.data['time_min_offset']
        self.check_integrity(verbose=verbose)
        if verbose:
            print(len(self), 'annotations imported.')

    def to_pamlab(self, outdir, outfile='PAMlab annotations.log', single_file=False):
        """
        Write data to 1 or several PAMlab files.

        Write annotations as .log files readable by the software PAMlab. Output
        files can be written in a single txt file or in several txt files (one
        per audio recording). In teh latter case, output file names are
        automatically generated based on the audio file's name and the name
        format required by PAMlab.

        Parameters
        ----------
        outdir : str
            Path of the output directory where the Raven files are written.
        outfile : str
            Name of teh output file. Only used is single_file is True. The
            default is 'PAMlab annotations.log'.
        single_file : bool, optional
            If set to True, writes a single output file with all annotations.
            The default is False.

        Returns
        -------
        None.

        """
        if single_file:
            annots = [self.data]
        else:
            annots = [pd.DataFrame(y)
                      for x, y in self.data.groupby(
                              'audio_file_name', as_index=False)]
        for annot in annots:
            annot.reset_index(inplace=True, drop=True)
            cols = ['fieldkey:', 'Soundfile', 'Channel', 'Sampling freq (Hz)',
                    'Latitude (deg)', 'Longitude (deg)', 'Recorder ID',
                    'Recorder depth', 'Start date and time (UTC)',
                    'Annotation date and time (local)', 'Recorder type',
                    'Deployment', 'Station', 'Operator', 'Left time (sec)',
                    'Right time (sec)', 'Top freq (Hz)', 'Bottom freq (Hz)',
                    'Species', 'Call type', 'rms SPL', 'SEL', '', '']
            outdf = pd.DataFrame({'fieldkey:': 0, 'Soundfile': 0, 'Channel': 0,
                                  'Sampling freq (Hz)': 0, 'Latitude (deg)': 0,
                                  'Longitude (deg)': 0, 'Recorder ID': 0,
                                  'Recorder depth': 0,
                                  'Start date and time (UTC)': 0,
                                  'Annotation date and time (local)': 0,
                                  'Recorder type': 0, 'Deployment': 0,
                                  'Station': 0, 'Operator': 0,
                                  'Left time (sec)': 0, 'Right time (sec)': 0,
                                  'Top freq (Hz)': 0, 'Bottom freq (Hz)': 0,
                                  'Species': '', 'Call type': '', 'rms SPL': 0,
                                  'SEL': 0, '': '', '': ''},
                                 index=list(range(annot.shape[0])))
            outdf['fieldkey:'] = 'an:'
            outdf['Species'] = annot['label_class']
            outdf['Call type'] = annot['label_subclass']
            outdf['Left time (sec)'] = annot['time_min_offset']
            outdf['Right time (sec)'] = annot['time_max_offset']
            outdf['Top freq (Hz)'] = annot['frequency_max']
            outdf['Bottom freq (Hz)'] = annot['frequency_min']
            outdf['rms SPL'] = annot['confidence']
            outdf['Operator'] = annot['operator_name']
            outdf['Channel'] = annot['audio_channel']
            outdf['Annotation date and time (local)'] = annot['entry_date']
            outdf['Sampling freq (Hz)'] = annot['audio_sampling_frequency']
            outdf['Recorder type'] = annot['recorder_type']
            outdf['Recorder ID'] = annot['recorder_SN']
            outdf['Recorder depth'] = annot['hydrophone_depth']
            outdf['Station'] = annot['location_name']
            outdf['Latitude (deg)'] = annot['location_lat']
            outdf['Longitude (deg)'] = annot['location_lon']
            outdf['Soundfile'] = [os.path.join(x, y) + z for x, y, z
                                  in zip(annot['audio_file_dir'],
                                         annot['audio_file_name'],
                                         annot['audio_file_extension']
                                         )
                                  ]
            outdf = outdf.fillna(0)
            if single_file:
                outfilename = os.path.join(outdir, outfile)
            else:
                outfilename = os.path.join(
                    outdir,
                    str(annot['audio_file_name'].iloc[0]))
                + str(annot['audio_file_extension'].iloc[0])
                + ' annotations.log'
            outdf.to_csv(outfilename,
                         sep='\t',
                         encoding='utf-8',
                         header=True,
                         columns=cols,
                         index=False)

    def from_parquet(self, file, verbose=False):
        """
        Import data from a Parquet file.

        Load annotations from a .parquet file. This format allows for fast and
        efficient data storage and access.

        Parameters
        ----------
        file : str
            Path of the input parquet file.

        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        self.data = pd.read_parquet(file)
        self.check_integrity(verbose=verbose)
        if verbose:
            print(len(self), 'annotations imported.')

    def to_parquet(self, file):
        """
        Write data to a Parquet file.

        Write annotations as .parquet file. This format allows for fast and
        efficient data storage and access.

        Parameters
        ----------
        file : str
            Path of the output directory where the parquet files is written.

        Returns
        -------
        None.

        """
        # make sure the HP SN column are strings
        self.data.hydrophone_SN = self.data.hydrophone_SN.astype(str)
        # save
        self.data.to_parquet(file,
                             coerce_timestamps='ms',
                             allow_truncated_timestamps=True)

    def from_netcdf(self, file, verbose=False):
        """
        Import data from a netcdf file.

        Load annotations from a .nc file. This format works well with xarray
        and Dask.

        Parameters
        ----------
        file : str
            Path of the nc file to import. Can be a str if importing a single
            file or entire folder. Needs to be a list if importing multiple
            files. If 'files' is a folder, all files in that folder ending with
            '.nc' will be imported.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if type(file) is str:
            if os.path.isdir(file):
                file = ecosound.core.tools.list_files(
                    file,
                    '.nc',
                    recursive=False,
                    case_sensitive=True,)
                if verbose:
                    print(len(file), 'files found.')
            else:
                file = [file]
        self.data = self._import_netcdf_files(file)
        self.check_integrity(verbose=verbose)
        if verbose:
            print(len(self), 'annotations imported.')

    def to_netcdf(self, file):
        """
        Write data to a netcdf file.

        Write annotations as .nc file. This format works well with xarray
        and Dask.

        Parameters
        ----------
        file : str
            Path of the output file (.nc) to be written.

        Returns
        -------
        None.

        """
        if file.endswith('.nc') is False:
            file = file + '.nc'
        self._enforce_dtypes()
        meas = self.data
        meas.set_index('time_min_date', drop=False, inplace=True)
        meas.index.name = 'date'
        dxr1 = meas.to_xarray()
        dxr1.attrs['datatype'] = 'Annotation'
        dxr1.to_netcdf(file, engine='netcdf4', format='NETCDF4')

    def insert_values(self, **kwargs):
        """
        Insert constant values for given Annotation fields.

        Fill in entire columns of the annotation dataframe with constant
        values. It is usefull for adding project related informations that may
        not be included in data imported from Raven or PAMlab files (e.g.,
        'location_lat', 'location_lon'). Values can be inserted for several
        annotations fields at a time by setting several keywords. This should
        only be used for filling in static values (i.e., not for variable
        values such as time/frequency boundaries of the annotations). Keywords
        must have the exact same name as the annotation field (see method
        .get_fields). For example: (location_lat=48.6, recorder_type='AMAR')

        Parameters
        ----------
        **kwargs : annotation filed name
            Keyword and value of the annotation field to fill in. Keywords must
            have the exact same name as the annotation field.

        Raises
        ------
        ValueError
            If keyword doesn't match any annotation field name.

        Returns
        -------
        None.

        """
        for key, value in kwargs.items():
            if key in self.data:
                self.data[key] = value
            else:
                raise ValueError('The annotation object has no field: '
                                 + str(key))

    def insert_metadata(self, deployment_info_file):
        """
        Insert metadata infortion to the annotation.

        Uses the Deployment_info_file to fill in the metadata of the annotation
        . The deployment_info_file must be created using the DeploymentInfo
        class from ecosound.core.metadata using DeploymentInfo.write_template.

        Parameters
        ----------
        deployment_info_file : str
            Csv file readable by ecosound.core.meta.DeploymentInfo.read(). It
            contains all the deployment metadata.

        Returns
        -------
        None.

        """
        dep_info = DeploymentInfo()
        dep_info.read(deployment_info_file)
        self.insert_values(UTC_offset=dep_info.data['UTC_offset'].values[0],
                           audio_channel=dep_info.data['audio_channel_number'].values[0],
                           audio_sampling_frequency=dep_info.data['sampling_frequency'].values[0],
                           audio_bit_depth=dep_info.data['bit_depth'].values[0],
                           mooring_platform_name = dep_info.data['mooring_platform_name'].values[0],
                           recorder_type=dep_info.data['recorder_type'].values[0],
                           recorder_SN=dep_info.data['recorder_SN'].values[0],
                           hydrophone_model=dep_info.data['hydrophone_model'].values[0],
                           hydrophone_SN=dep_info.data['hydrophone_SN'].values[0],
                           hydrophone_depth=dep_info.data['hydrophone_depth'].values[0],
                           location_name=dep_info.data['location_name'].values[0],
                           location_lat=dep_info.data['location_lat'].values[0],
                           location_lon=dep_info.data['location_lon'].values[0],
                           location_water_depth=dep_info.data['location_water_depth'].values[0],
                           deployment_ID=dep_info.data['deployment_ID'].values[0],
                          )

    def filter_overlap_with(self, annot, freq_ovp=True,dur_factor_max=None,dur_factor_min=None,ovlp_ratio_min=None,remove_duplicates=False,inherit_metadata=False,filter_deploymentID=True, inplace=False):
        """
        Filter overalaping annotations.

        Only keep annotations that overlap in time and/or frequency with the
        annotation object "annot".

        Parameters
        ----------
        annot : ecosound.annotation.Annotation object
            Annotation object used to filter the current annotations.
        freq_ovp : bool, optional
            If set to True, filters not only annotations that overlap in time
            but also overlap in frequency. The default is True.
        dur_factor_max : float, optional
            Constraint dictating the maximum duration overlapped
            annotations must not exceed in order to be "kept". Any annotations
            whose duration exceed dur_factor_max*annot.duration are discareded,
            even if they overlap in time/frequency. If set to None, no maximum
            duration constraints are applied. The default is None.
        dur_factor_min : float, optional
            Constraint dictating the minimum duration overlapped
            annotations must exceed in order to be "kept". Any annotations
            whose duration does not exceed dur_factor_min*annot.duration are
            discareded, even if they overlap in time/frequency. If set to None,
            no minimum duration constraints are applied. The default is None.
        ovlp_ratio_min : float, optional
            Constraint dictating the minimum amount (percentage) of overlap in
            time annotations must have in order to be "kept". If set to None,
            no minimum time overlap constraints are applied. The default is
            None.
        remove_duplicates : bool, optional
            If set to True, only selects a single annotation overlaping with
            annotations from the annot object. This is relevant only if several
            annotations overlap with an annotation from the annot object. The
            default is False.
        inherit_metadata : bool, optional
            If set to True, the filtered annotations inherit all the metadata
            information from the matched annotations in the annot object. It
            includes 'label_class', 'label_subclass', 'mooring_platform_name',
            'recorder_type', 'recorder_SN', 'hydrophone_model', 'hydrophone_SN'
            , 'hydrophone_depth', 'location_name', 'location_lat',
            'location_lon', 'location_water_depth', and 'deployment_ID'. The
            default is False.
        filter_deploymentID : bool, optional
            If set to False, doesn't use the deploymentID to match annotations
            together but just the frequency and time offset boundaries of the
            annotations. The default is True.
        inplace : bool, optional
            If set to True, updates the urrent object with the filter results.
            The default is False.

        Returns
        -------
        out_object : ecosound.annotation.Annotation
            Filtered Annotation object.

        """
        stack = []
        det = self.data
        for index, an in annot.data.iterrows():  # for each annotation
            # restrict to the specific deploymnetID of the annotation if file names are not unique
            if filter_deploymentID:
                df = det[det.deployment_ID == an.deployment_ID]
            else:
                df = det
            ## filter detections to same file and deployment ID as the current annotation
            df = df[df.audio_file_name == an.audio_file_name]
            ## check overlap in time first
            if len(df) > 0:
                df = df[((df.time_min_offset <= an.time_min_offset) & (df.time_max_offset >= an.time_max_offset)) |  # 1- annot inside detec
                         ((df.time_min_offset >= an.time_min_offset) & (df.time_max_offset <= an.time_max_offset)) |  # 2- detec inside annot
                         ((df.time_min_offset < an.time_min_offset) & (df.time_max_offset < an.time_max_offset) & (df.time_max_offset > an.time_min_offset)) | # 3- only the end of the detec overlaps with annot
                         ((df.time_min_offset > an.time_min_offset) & (df.time_min_offset < an.time_max_offset) & (df.time_max_offset > an.time_max_offset)) # 4- only the begining of the detec overlaps with annot
                          ]
            # then looks at frequency overlap. Can be turned off if freq bounds are not reliable
            if (len(df) > 0) & freq_ovp:
                df = df[((df.frequency_min <= an.frequency_min) & (df.frequency_max >= an.frequency_max)) | # 1- annot inside detec
                        ((df.frequency_min >= an.frequency_min) & (df.frequency_max <= an.frequency_max)) | # 2- detec inside annot
                        ((df.frequency_min < an.frequency_min) & (df.frequency_max < an.frequency_max) & (df.frequency_max > an.frequency_min)) | # 3- only the top of the detec overlaps with annot
                        ((df.frequency_min > an.frequency_min) & (df.frequency_min < an.frequency_max) & (df.frequency_max > an.frequency_max)) # 4- only the bottom of the detec overlaps with annot
                        ]
            # discard if durations are too different
            if (len(df) > 0) & (dur_factor_max is not None):
                df = df[df.duration < an.duration*dur_factor_max]
            if (len(df) > 0) & (dur_factor_min is not None):
                df = df[df.duration > an.duration*dur_factor_min]

            # discard if they don't overlap enough
            if (len(df) > 0) & (ovlp_ratio_min is not None):
                df_ovlp = (df['time_max_offset'].apply(lambda x: min(x,an.time_max_offset)) - df['time_min_offset'].apply(lambda x: max(x,an.time_min_offset))) / an.duration
                df = df[df_ovlp>=ovlp_ratio_min]
                df_ovlp = df_ovlp[df_ovlp>=ovlp_ratio_min]

            if (len(df) > 1) & remove_duplicates:
                try:
                    df = df.iloc[[df_ovlp.values.argmax()]] # pick teh one with max time overlap
                except:
                    print('asas')

            if len(df) > 0:
                if inherit_metadata:
                    df['mooring_platform_name'] = an['mooring_platform_name']
                    df['recorder_type'] = an['recorder_type']
                    df['recorder_SN'] = an['recorder_SN']
                    df['hydrophone_model'] = an['hydrophone_model']
                    df['hydrophone_SN'] = an['hydrophone_SN']
                    df['hydrophone_depth'] = an['hydrophone_depth']
                    df['location_name'] = an['location_name']
                    df['location_lat'] = an['location_lat']
                    df['location_lon'] = an['location_lon']
                    df['location_water_depth'] = an['location_water_depth']
                    df['deployment_ID'] = an['deployment_ID']
                    df['label_class'] = an['label_class']
                    df['label_subclass'] = an['label_subclass']
                stack.append(df)
        ovlp = pd.concat(stack, ignore_index=True)
        if inplace:
            self.data = ovlp
            self.check_integrity()
            out_object = None
        else:
            out_object = copy.copy(self)
            out_object.data = ovlp
            out_object.check_integrity()
        return out_object

    def get_labels_class(self):
        """
        Get all the unique class labels of the annotations.

        Returns
        -------
        classes : list
            List of unique class labels.

        """
        if len(self.data) > 0:
            classes = list(self.data['label_class'].unique())
        else:
            classes = []
        return classes

    def get_labels_subclass(self):
        """
        Get all the unique subclass labels of the annotations.

        Returns
        -------
        classes : list
            List of unique subclass labels.

        """
        if len(self.data) > 0:
            subclasses = list(self.data['label_subclass'].unique())
        else:
            subclasses = []
        return subclasses

    def get_fields(self):
        """
        Get all the annotations fields.

        Returns
        -------
        classes : list
            List of annotation fields.

        """
        return list(self.data.columns)

    def summary(self, rows='deployment_ID', columns='label_class'):
        """
        Produce a summary table of the number of annotations.

        Create a pivot table summarizing the number of annotations for each
        deployment and each label class. The optional arguments 'rows' and
        'columns' can be used to change the fields of the annotations to be
        displayed in the table.

        Parameters
        ----------
        rows : 'str', optional
            Name of the annotation field for the rows of the table. The default
            is 'deployment_ID'.
        columns : 'str', optional
            Name of the annotation field for the columns of the table. The
            default default is 'label_class'.

        Returns
        -------
        summary : pandas DataFrame
            Pivot table with the number of annotations in each category.

        """
        summary = self.data.pivot_table(index=rows,
                                        columns=columns,
                                        aggfunc='size',
                                        fill_value=0)
        # Add a "Total" row and column
        summary.loc['Total'] = summary.sum()
        summary['Total'] = summary.sum(axis=1)
        return summary

    def _enforce_dtypes(self):
        self.data = self.data.astype({
            'uuid': 'str',
            'from_detector': 'bool',  # True, False
            'software_name': 'str',
            'software_version': 'str',
            'operator_name': 'str',
            'UTC_offset': 'float',
            'entry_date': 'datetime64[ns]',
            'audio_channel': 'int',
            'audio_file_name': 'str',
            'audio_file_dir': 'str',
            'audio_file_extension': 'str',
            'audio_file_start_date': 'datetime64[ns]',
            'audio_sampling_frequency': 'int',
            'audio_bit_depth': 'int',
            'mooring_platform_name': 'str',
            'recorder_type': 'str',
            'recorder_SN': 'str',
            'hydrophone_model': 'str',
            'hydrophone_SN': 'str',
            'hydrophone_depth': 'float',
            'location_name': 'str',
            'location_lat': 'float',
            'location_lon': 'float',
            'location_water_depth': 'float',
            'deployment_ID': 'str',
            'frequency_min': 'float',
            'frequency_max': 'float',
            'time_min_offset': 'float',
            'time_max_offset': 'float',
            'time_min_date': 'datetime64[ns]',
            'time_max_date': 'datetime64[ns]',
            'duration': 'float',
            'label_class': 'str',
            'label_subclass': 'str',
            'confidence': 'float',
            })

    @staticmethod
    @ecosound.core.decorators.listinput
    def _import_csv_files(files):
        """Import one or several text files with header to a Panda datafrane."""
        assert type(files) in (str, list), "Input must be of type str (single \
            file) or list (multiple files)"
        # Import all files to a dataframe
        for idx, file in enumerate(files):
            # Extract header first due to formating issues in PAMlab files
            header = pd.read_csv(file,
                                 delimiter='\t',
                                 header=None,
                                 nrows=1)
            headerLength = header.shape[1]
            # Get all data and only keep values correpsonding to header labels
            tmp = pd.read_csv(file,
                              delimiter='\t',
                              header=None,
                              skiprows=1,
                              na_values=None)
            tmp = tmp.iloc[:, 0:headerLength]
            # Put header back
            tmp = tmp.set_axis(list(header.values[0]), axis=1, inplace=False)
            if idx == 0:
                data = tmp
            else:
                data = pd.concat([data, tmp], ignore_index=True, sort=False)
        return data

    def _import_netcdf_files(self, files):
        """Import one or several netcdf files to a Panda datafrane."""
        assert type(files) in (str, list), "Input must be of type str (single \
            file or directory) or list (multiple files)"
        # Import all files to a dataframe
        tmp = []
        for idx, file in enumerate(files):
            dxr = xr.open_dataset(file)
            if dxr.attrs['datatype'] == 'Annotation':
                tmp2 = dxr.to_dataframe()
                tmp2.reset_index(inplace=True)
            elif dxr.attrs['datatype'] == 'Measurement':
                tmp2 = dxr.to_dataframe()
                tmp2.reset_index(inplace=True)
                tmp2 = tmp2[self.get_fields()]
                warnings.warn('Importing Measurement data as Annotation >> Not all Measurement data are loaded.')
            else:
                raise ValueError(file + 'Not an Annotation file.')
            tmp.append(tmp2)

        data = pd.concat(tmp, ignore_index=True, sort=False)
        data.reset_index(inplace=True, drop=True)
        return data

    def __add__(self, other):
        """Concatenate data from several annotation objects."""
        assert type(other) is ecosound.core.annotation.Annotation, "Object type not \
            supported. Can only concatenate Annotation objects together."
        self._enforce_dtypes()
        other._enforce_dtypes()
        self.data = pd.concat([self.data, other.data],
                              ignore_index=True,
                              sort=False)
        return self

    def __repr__(self):
        """Return the type of object."""
        return (f'{self.__class__.__name__} object ('
                f'{len(self.data)})')

    def __str__(self):
        """Return string when used with print of str."""
        return f'{len(self.data)} annotation(s)'

    def __len__(self):
        """Return number of annotations."""
        return len(self.data)



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 16 15:51:43 2020

@author: xavier.mouy
"""
import pickle
model_filename = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Classification\LDA_model.sav'
import matplotlib.pyplot as plt
import numpy as np

# load the model from disk
#loaded_model = pickle.load(open(model_filename, 'rb'))
#result = loaded_model.score(X_test, Y_test)
#print(result)

y=np.random.rand(100,1)
x=range(0,100,1)
fig, ax = plt.subplots(1,1)
ax.plot(x,y)

#style = dict(size=10, color='gray')
bbox_props = dict(boxstyle="square", fc="w", ec="w", alpha=0.5)
ax.text(1, 1, "New Year's Day", size=5, bbox=bbox_props)


# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 21:21:40 2020

@author: xavier.mouy
"""
import json
import re
from datetime import datetime
import ecosound.core.decorators
import numpy as np
from scipy import interpolate
import os, sys
from numba import njit
import pkg_resources

def read_json(file):
    """Load JSON file as dict."""
    with open(file, "r") as read_file:
        data = json.load(read_file)
    return data


@ecosound.core.decorators.listinput
def filename_to_datetime(files):
    """Extract date from a list of str of filenames."""
    current_dir = os.path.dirname(os.path.realpath(__file__))
    patterns = read_json(os.path.join(current_dir, r'timestamp_formats.json'))
    
    #stream = pkg_resources.resource_stream(__name__, 'core/timestamp_formats.json')
    #patterns = read_json(os.path.join(stream)
    
    regex_string = '|'.join([pattern['string_pattern'] for pattern in patterns])
    time_formats = [pattern['time_format'] for pattern in patterns]
    timestamps = [None] * len(files)
    p = re.compile(regex_string)
    for idx, file in enumerate(files):
        datestr = p.search(file)
        for time_format in time_formats:
            ok_flag = False
            try:
                timestamps[idx] = datetime.strptime(datestr[0], time_format)
                ok_flag = True
            except:
                ok_flag = False
            if ok_flag is True:
                break
        if ok_flag is False:
            raise ValueError('Time format not recognized:' + file)
    return timestamps

#@njit
def normalize_vector(vec):
    """ 
    Normalize amplitude of vector.
    """
    # vec = vec+abs(min(vec))
    # normVec = vec/max(vec)
    # normVec = (normVec - 0.5)*2
    vec = vec - np.mean(vec)
    normVec = vec/max(vec)
    return normVec

#@njit
def tighten_signal_limits(signal, energy_percentage):
    """
    Tighten signal limits

    Redefine start and stop samples to have "energy_percentage" of the original
    signal. Returns a list with the new start and stop sample indices.

    """
    cumul_energy = np.cumsum(np.square(signal))
    cumul_energy = cumul_energy/max(cumul_energy)
    percentage_begining = (1-(energy_percentage/100))/2
    percentage_end = 1 - percentage_begining
    chunk = [np.nonzero(cumul_energy > percentage_begining)[0][0], 
             np.nonzero(cumul_energy > percentage_end)[0][0]]
    return chunk

def resample_1D_array(x, y, resolution, kind='linear'):
    """
    Interpolate values of coordinates x and y with a given resolution.
    Default uisn linear interpolation.
    """
    f = interpolate.interp1d(x, y, kind=kind, fill_value='extrapolate')   
    xnew = np.arange(x[0], x[-1]+resolution, resolution)
    ynew = f(xnew)
    return xnew, ynew

@njit
def entropy(array_1d, apply_square=False):
        """ 
        Aggregate (SHannon's) entropy as defined in the Raven manual
        apply_square = True, suqares the array value before calculation.
        """
        if apply_square:
            array_1d = np.square(array_1d)
        values_sum = np.sum(array_1d)
        H = 0
        for value in array_1d:
            ratio = (value/values_sum)
            if ratio <= 0:
                H += 0
            else:
                H += ratio*np.log2(ratio)
        return H

@njit
def derivative_1d(array, order=1):
    """
    Derivative of order "order" of a 1D array. Subtract the element i+1 to i.
    If order > 1, the process is repeated iteratively "order" times. Note that
    the resulting array is shorter than the original array by "order" elements.

    """
    for n in range(0, order, 1):
        array = np.subtract(array[1:], array[0:-1])
    return array

def list_files(indir, suffix, case_sensitive=True, recursive=False):
    """
    List files in folder whose name ends with a given suffix/extension.
    
    Parameters
    ----------
    indir : str
        Path of the folder to search.
    suffix : str
        Suffix of the filename. 
    case_sensitive : bool, optional
        If set to True, search using case sensitive filenames. The default is
        True.
    recursive : bool, optional
        If set to True, search in parent folder but also in all its subfolders.
        The default is False.

    Returns
    -------
    files_list : list
        List of strings with full path of files found.

    """
    if os.path.isdir(indir):
        files_list = []
        if case_sensitive is False:
            suffix = suffix.lower()
        if recursive:  # scans subfolders recursively
            for root, dirs, files in os.walk(indir):
                for file in files:
                    if case_sensitive is False:
                        file = file.lower()
                    if file.endswith(suffix):
                        files_list.append(os.path.join(root, file))
                        #print(os.path.join(root, file))
        else:  # only scans parent folder
            for file in os.listdir(indir):
                if case_sensitive is False:
                    file = file.lower()
                if file.endswith(suffix):
                    files_list.append(os.path.join(indir, file))
                    #print(os.path.join(indir, file))
    return files_list
            


# -*- coding: utf-8 -*-
"""
Created on Sun Mar  1 21:26:33 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..") # Adds higher directory to python modules path.
from scipy import signal
import matplotlib.pyplot as plt
import numpy as np
import time
from dask import delayed, compute
import dask.bag as db
from ecosound.core.audiotools import Sound

# from dask.distributed import Client, LocalCluster
# cluster = LocalCluster()
# client = Client(cluster,processes=False)




# Using scipy #########################################
def spectro_numpy(sig, fs, frame_samp, overlap_samp, fft_samp):
    tic = time.perf_counter()
    axis_frequencies, axis_times, spectrogram = signal.spectrogram(sig,
                                                                    fs=fs,
                                                                    window=signal.hann(frame_samp),
                                                                    noverlap=overlap_samp,
                                                                    nfft=fft_samp,
                                                                    scaling='spectrum')
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")
    return spectrogram

def getFFT(sig,nfft):
    s = np.fft.fft(sig, nfft)
    s2 = abs(s)

    return s2

def calc_spectrogram(sig, win, starts, stops, fft_samp):
    fnyq = int(np.round(fft_samp/2))
    spectro = np.empty((fnyq,len(starts))) # the default 
    idx=0
    for start, stop in zip(starts, stops):
        s = sig[start:stop]*win
        Spectrum = np.fft.fft(s, fft_samp)
        Spectrum = abs(Spectrum) # amplitude
        Spectrum = Spectrum*2
        #Spectrum = Spectrum**2
        #ts_window = getFFT(sig[start:stop],fft_samp)
        spectro[:,idx] = Spectrum[0:fnyq]
        idx+=1
    return spectro

def spectro_loop(sig, fs, frame_samp, overlap_samp, fft_samp):
    tic = time.perf_counter()
    starts = np.arange(0,len(sig),frame_samp-overlap_samp,dtype=int)
    starts = starts[starts + frame_samp < len(sig)]
    xns = []
    for start in starts:
        # short term discrete fourier transform
        ts_window = getFFT(sig[start:start + frame_samp],fft_samp)
        xns.append(ts_window)
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")
    return xns

def spectro_loop_dask(sig, fs, frame_samp,overlap_samp,fft_samp):
    tic = time.perf_counter()
    starts = np.arange(0,len(sig), frame_samp - overlap_samp, dtype=int)
    starts = starts[starts + frame_samp < len(sig)]
    xns = []
    for start in starts:
        sig2 = sig[start:start + frame_samp]
        # short term discrete fourier transform
        ts_window = delayed(getFFT)(sig2, fft_samp)
        xns.append(ts_window)
    compute(xns)
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")
    return xns

def spectro_loop_dask2(sig, fs, frame_samp,overlap_samp,fft_samp, dask=False, dask_chunks=40):
    #dask_chunks = 40
    tic = time.perf_counter()
    step = frame_samp - overlap_samp
    starts = np.arange(0,len(sig)-frame_samp, step, dtype=int)
    stops = starts + frame_samp
    start_chunks = np.array_split(starts,dask_chunks)
    stop_chunks = np.array_split(stops,dask_chunks)
    win = np.hanning(frame_samp)
    spectrogram = []
    idx=0
    for start_chunk, stop_chunk in zip(start_chunks, stop_chunks):
        sig_chunk = sig[start_chunk[0]:stop_chunk[-1]]
        chunk_size = len(start_chunk)
        if dask:
            spectro_chunk = delayed(calc_spectrogram)(sig_chunk,win,
                                       start_chunk-start_chunk[0],
                                       stop_chunk-start_chunk[0],
                                       fft_samp)
        else:
            spectro_chunk = calc_spectrogram(sig_chunk,win,
                            start_chunk-start_chunk[0],
                            stop_chunk-start_chunk[0],
                            fft_samp)
        spectrogram.append(spectro_chunk)
        idx += chunk_size
    if dask:
        spectrogram = compute(spectrogram)
        spectrogram = np.concatenate(spectrogram[0][:], axis=1)
    else:
        spectrogram = np.concatenate(spectrogram[:], axis=1)
    axis_times = starts/fs
    freq_resolution = fs/fft_samp
    axis_frequencies = np.arange(0,fs/2,freq_resolution)
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")
    return axis_frequencies, axis_times, spectrogram
if __name__ == '__main__':

    # # Create random signal
    # fs = 48000
    # sig_dur = 60#1800
    # sig = np.random.rand(sig_dur*fs)

    single_channel_file = r"../ecosound/resources/67674121.181018013806.wav"

    t1 = 24
    t2 = 120
    sound = Sound(single_channel_file)
    #sound.read(channel=0, chunk=[t1, t2], unit='sec')
    sound.read(channel=0)
    fs = sound.file_sampling_frequency
    sig = sound.waveform
    sig = sig-np.mean(sig)


    # Calculates  spectrogram
    frame_samp = 3000
    overlap_samp = 2500
    fft_samp = 4096

    #S1 = spectro_numpy(sig, fs, frame_samp, overlap_samp, fft_samp)
    #S2 = spectro_loop(sig, fs, frame_samp, overlap_samp, fft_samp)
    #S3 = spectro_loop_dask(sig, fs, frame_samp, overlap_samp, fft_samp)
    F, T, S3 = spectro_loop_dask2(sig,
                                  fs,
                                  frame_samp,
                                  overlap_samp,
                                  fft_samp,
                                  dask=True,
                                  dask_chunks=50,
                                  )
    
    
    # fig, ax = plt.subplots(figsize=(16,4), sharex=True)
    # im = ax.pcolormesh(S1, cmap='jet',vmin = np.percentile(S1,50), vmax= np.percentile(S1,99.9))

    fig, ax = plt.subplots(figsize=(16,4), sharex=True)
    im = ax.pcolormesh(T, F, S3, cmap='jet',vmin = np.percentile(S3,50), vmax= np.percentile(S3,99.9))



# -*- coding: utf-8 -*-
"""
Created on Sun Mar  1 21:26:33 2020

@author: xavier.mouy
"""
# import sys
# sys.path.append("..") # Adds higher directory to python modules path.
# from ecosound.core.audiotools import Sound
# from scipy import signal
# import matplotlib.pyplot as plt
# import numpy as np
# import time

# from dask import delayed

# single_channel_file = r"../ecosound/resources/67674121.181018013806.wav"

# # Spectrogram parameters
# frame = 3000
# nfft = 4096
# step = 500
# #ovlp = 2500
# fmin = 0 
# fmax = 1000
# window_type = 'hann'

# # start and stop time of wavfile to analyze
# t1 = 24
# t2 = 40
# ## ###########################################################################


# # load audio data
# sound = Sound(single_channel_file)
# sound.read(channel=0, chunk=[t1, t2], unit='sec')

# # Calculates  spectrogram
# frame_samp = 1024
# overlap_samp = 512
# fft_samp = 1024

## Using scipy #########################################
# tic = time.perf_counter()
# axis_frequencies, axis_times, spectrogram = signal.spectrogram(sound.waveform,
#                                                                fs=sound.waveform_sampling_frequency,
#                                                                window=signal.hann(frame_samp),
#                                                                noverlap=overlap_samp,
#                                                                nfft=fft_samp,
#                                                                scaling='spectrum')

# toc = time.perf_counter()
# print(f"Dexecuted in {toc - tic:0.4f} seconds")
# #fig, ax = plt.subplots(figsize=(16,4), sharex=True)
# #im = ax.pcolormesh(axis_times,axis_frequencies,spectrogram, cmap='jet',vmin = np.percentile(spectrogram,50), vmax= np.percentile(spectrogram,99.9))


# ## Using Dask #########################################
# from dask import delayed
# from dask.distributed import Client, progress
# client = Client(n_workers=4)


# def getFFT(sig):
#     return abs(np.fft.fft(sig))

# def crop(x,fft_samp):
#     return x[0:int(fft_samp/2)]

# tic = time.perf_counter()
# starts = np.arange(0,len(sound.waveform),fft_samp-overlap_samp,dtype=int)
# starts = starts[starts + fft_samp < len(sound.waveform)]
# xns = []
# for start in starts:
#     # short term discrete fourier transform
#     ts_window = delayed(getFFT)(sound.waveform[start:start + fft_samp])
#     ts_window2 = delayed(crop)(ts_window,fft_samp)
#     xns.append(ts_window2)

# specX = delayed(np.array(xns).T)
# specX.visualize()
# # #spec = delayed(10*np.log10(specX))
# # result = specX.compute()

# # toc = time.perf_counter()
# # print(f"Dexecuted in {toc - tic:0.4f} seconds")

# #fig2, ax2 = plt.subplots(figsize=(16,4), sharex=True)
# #im = ax2.pcolormesh(spec, cmap='jet',vmin = np.percentile(spec,50), vmax= np.percentile(spec,99.9))

import time
from dask.distributed import Client
from dask import delayed
from time import sleep
    
## Using Dask #########################################

def inc(x):
        sleep(1)
        print('inc')
        return x + 1

def add(x, y):
    sleep(1)
    print('add')
    return x + y
    
def test():

    client = Client(n_workers=10)
    
    tic = time.perf_counter()
    x = delayed(inc)(1)
    y = delayed(inc)(2)
    z = delayed(add)(x, y)
    total = z.compute()
    print(total)
    toc = time.perf_counter()
    print(f"Downloaded the tutorial in {toc - tic:0.4f} seconds")
    client.close()

if __name__ == '__main__':
    test()




# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 15:28:29 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement
from ecosound.core.annotation import Annotation


# PAMlab_files = []
# Raven_files = []
# Raven_files.append(r".\resources\AMAR173.4.20190916T061248Z.Table.1.selections.txt")
# annot1 = Annotation()
# annot1.from_raven(Raven_files, verbose=False)
# print(len(annot1))

# Raven_files = []
# Raven_files.append(r".\resources\67674121.181018013806.Table.1.selections.txt")
# annot2 = Annotation()
# annot2.from_raven(Raven_files, verbose=False)
# print(len(annot2))

# PAMlab_files = []
# PAMlab_files.append(r"..\ecosound\resources\JASCOAMARHYDROPHONE742_20140913T084017.774Z.wav annotations.log")
# annot3 = Annotation()
# annot3.from_pamlab(PAMlab_files, verbose=False)
# print(len(annot3))

# annot = annot1 + annot2 + annot3
# print(len(annot))

# print(annot.get_fields())
# annot.insert_values(operator_name='Xavier Mouy')


# ## netcdf 1 file only
# netcdf_files = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\DFO_snake-island_rca-in_20181017\Annotations_dataset_SI-RCAIn-20181017.nc"
# annot4 = Annotation()
# annot4.from_netcdf(netcdf_files, verbose=True)
# print(len(annot4))


# # ## netcdf several files
# netcdf_files = []
# netcdf_files.append(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\DFO_snake-island_rca-in_20181017\Annotations_dataset_SI-RCAIn-20181017.nc")
# netcdf_files.append(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\DFO_snake-island_rca-out_20181015\Annotations_dataset_SI-RCAOut-20181015.nc")
# netcdf_files.append(r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\ONC_delta-node_2014\Annotations_dataset_ONC-Delta-2014.nc")
# annot4 = Annotation()
# annot4.from_netcdf(netcdf_files, verbose=True)
# print(len(annot4))

# # ## netcdf folder
# netcdf_files = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test'
# annot4 = Annotation()
# annot4.from_netcdf(netcdf_files, verbose=True)
# print(len(annot4))

# # ## netcdf folder from Measurements
# netcdf_files = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test2'
# annot4 = Annotation()
# annot4.from_netcdf(netcdf_files, verbose=True)
# print(len(annot4))

# ## netcdf folder from Measurements
netcdf_files = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\test2'
annot4 = Measurement()
annot4.from_netcdf(netcdf_files, verbose=True)
print(len(annot4))

# import xarray as xr
# d=annot3.data
# index = range(0,len(d),1)
# d['index']=index
# #d = d.set_index(['index','entry_date', 'frequency_min','label_class'])
# d = d.set_index(['index'])

# data = d.to_xarray()

# data2=data.sel(index=0)


# -*- coding: utf-8 -*-
"""
Created on Thu May 25 09:54:11 2017

-------------------------
TODO:

    - fix plot of timeseries
    - plotDetectionFunction with unit='samp' not implemented yet
    - 
--------------------------

@author: xavier
"""
import pandas as pd
import numpy as np
import scipy.signal as spsig
import matplotlib.pyplot as plt
import os
from numba import jit

class KurtosisDetector:

    def __init__(self, SoundObj, Kurtframe_sec, Kurtth, Kurtdelta_sec):        
        self.Kurtframe_sec = Kurtframe_sec
        self.Kurtth = Kurtth
        self.Kurtdelta_sec = Kurtdelta_sec
        self.SoundObj = SoundObj
        self.Kurtframe_samp = int(self.Kurtframe_sec*SoundObj.getSamplingFrequencyHz())
        self.Kurtdelta_samp = int(self.Kurtdelta_sec*SoundObj.getSamplingFrequencyHz())

    def run(self):
        Kurt = self.calcKurtosis()
        Peaks = findPeaks(Kurt, self.Kurtth, self.Kurtdelta_sec)
        detec = Detections()
        detec.output['fileName'] = self.SoundObj.getFileName()
        detec.output['filePath'] = self.SoundObj.getFilePath()
        detec.output['fileExtension'] = self.SoundObj.getFileExtension()
        detec.output['startTimeSec'] = Peaks['Tsec']
        detec.output['stopTimeSec'] = Peaks['Tsec']
        detec.output['startTimeSamp'] = Peaks['Tsamp']
        detec.output['stopTimeSamp'] = Peaks['Tsamp']
        detec.output['confidence'] = Peaks['Value']
        detec.output['detectorName'] = 'KurtosisDetector'
        detec.output['type'] = 'detec'
        detec.output['channel'] = self.SoundObj.getSelectedChannel()
        detec.output['fileDurationSec'] = self.SoundObj.getFileDur_sec()
        self.detecFunction = Kurt
        self.detections = detec.output
        return self.detections

    @jit
    def calcKurtosis(self):
        sig = self.SoundObj.getWaveform()
        K = np.array(sig)
        Tidx = np.array([])
        ss = pd.Series(sig)
        K = ss.rolling(window=self.Kurtframe_samp, center=False).kurt()
        K.fillna(0, inplace=True)
        Tidx = np.arange(0, len(K), 1)
        Tsec = Tidx/self.SoundObj.getSamplingFrequencyHz()
        Kurt = pd.DataFrame({'Value': K, 'Tsec': Tsec, 'Tsamp': Tidx})
        return Kurt
    
    def plot(self, displayDetections=True, unit='sec', newFig=True):
        if displayDetections:
            detections = self.detections
        else:
            detections = []        
        ylabel = 'Kurtosis'
        fig = Figure(self.detecFunction, detections=detections, ylabel=ylabel)
        fig.plotDetectionFunction(unit='sec', newFig=newFig)


class Detections:

    def __init__(self):
        data = pd.DataFrame({
            'type': 'detec',
            'channel': [],
            'detectorName': [],
            'fileName': [],
            'filePath': [],
            'fileExtension': [],
            'startTimeSec': [],
            'stopTimeSec': [],
            'startTimeSamp': [],
            'stopTimeSamp': [],
            'freqMinHz': [],
            'freqMaxHz': [],
            'confidence': [],
            'species': [],
            'call': [],
            'fileDurationSec': []
            })
        self.output = data

    def save2Pamlab(self, outdir):
        cols = ['fieldkey:', 'Soundfile', 'Channel', 'Sampling freq (Hz)', 'Latitude (deg)', 'Longitude (deg)', 'Recorder ID', 'Recorder depth', 'Start date and time (UTC)', 'Annotation date and time (local)', 'Recorder type', 'Deployment', 'Station', 'Operator', 'Left time (sec)', 'Right time (sec)', 'Top freq (Hz)', 'Bottom freq (Hz)', 'Species', 'Call type', 'rms SPL', 'SEL', '', '']
        annot = pd.DataFrame({'fieldkey:': 0, 'Soundfile': 0, 'Channel': 0, 'Sampling freq (Hz)': 0, 'Latitude (deg)': 0, 'Longitude (deg)': 0, 'Recorder ID': 0, 'Recorder depth': 0, 'Start date and time (UTC)': 0, 'Annotation date and time (local)': 0, 'Recorder type': 0, 'Deployment': 0, 'Station': 0, 'Operator': 0, 'Left time (sec)': 0, 'Right time (sec)': 0, 'Top freq (Hz)': 0, 'Bottom freq (Hz)': 0, 'Species': 0, 'Call type': 0, 'rms SPL': 0, 'SEL': 0, '': 0, '': 0}, index=list(range(self.output.shape[0])))    
        annot['fieldkey:'] = 'an:'
        annot['Species'] = self.output['species']
        annot['Call type'] = self.output['call']
        annot['Left time (sec)'] = self.output['startTimeSec']
        annot['Right time (sec)'] = self.output['stopTimeSec']
        annot['Top freq (Hz)'] = self.output['freqMaxHz']
        annot['Bottom freq (Hz)'] = self.output['freqMinHz']
        annot['rms SPL'] = self.output['confidence']
        annot['Operator'] =self.output['detectorName']
        annot['Channel'] =self.output['channel']
        if len(self.output.fileName) == 0:
            annot['Soundfile'] = os.path.join(str(self.output.filePath[0]), str(self.output.fileName[0])) + str(self.output.fileExtension[0])
        else:
            filenames=[]
            for i in range(0,len(self.output.fileName)):
                filenames.append(os.path.join(str(self.output.filePath[i]), str(self.output.fileName[i])) + str(self.output.fileExtension[i]))
            annot['Soundfile'] = filenames             
        annot.to_csv(os.path.join(outdir, str(self.output.fileName[0])) + str(self.output.fileExtension[0]) + ' chan' +  str(self.output.channel[0]) + ' annotations.log', sep='\t', encoding='utf-8', header=True, columns=cols, index=False)
                
    
    def save2Raven(self, outdir):      
        cols = ['Selection', 'View', 'Channel', 'Begin Time (s)', 'End Time (s)', 'Low Freq (Hz)', 'High Freq (Hz)', 'Begin Path', 'File Offset (s)', 'Begin File', 'File Duration (s)', 'Species', 'Sound type', 'Detector', 'Distortion (DTW)']
        annot = pd.DataFrame({'Selection': 0, 'View': 0, 'Channel': 0, 'Begin Time (s)': 0, 'End Time (s)': 0, 'Low Freq (Hz)': 0, 'High Freq (Hz)': 0, 'Begin Path': 0, 'File Offset (s)': 0, 'Begin File': 0, 'File Duration (s)': 0, 'Species': 0, 'Sound type': 0, 'Detector': 0, 'Distortion (DTW)': 0}, index=list(range(self.output.shape[0])))    
        annot['Selection'] = range(1,self.output.shape[0]+1)
        annot['View'] = 'Spectrogram 1'
        annot['Channel'] = self.output['channel'] + 1
        annot['Begin Time (s)'] = self.output['startTimeSec']
        annot['End Time (s)'] = self.output['stopTimeSec']
        annot['Low Freq (Hz)'] = self.output['freqMinHz']
        annot['High Freq (Hz)'] = self.output['freqMaxHz']
        annot['Begin Path'] = os.path.join(str(self.output.filePath[0]), str(self.output.fileName[0])) + str(self.output.fileExtension[0])
        annot['File Offset (s)'] = self.output['startTimeSec']
        annot['Begin File'] = str(self.output.fileName[0]) + str(self.output.fileExtension[0])
        annot['File Duration (s)'] = self.output['fileDurationSec']
        annot['Species'] = self.output['species']
        annot['Sound type'] = self.output['call']
        annot['Detector'] =self.output['detectorName']
        annot['Distortion (DTW)'] = self.output['confidence']
        annot.to_csv(os.path.join(outdir, str(self.output.fileName[0])) + str(self.output.fileExtension[0]) + '.chan' +  str(self.output.channel[0]) + '.Table.1.selections.txt', sep='\t', encoding='utf-8', header=True, columns=cols, index=False)

class Figure:
    def __init__(self, detecFunction, detections=[], ylabel='Detection value'):
        self.axis_t_sec = detecFunction['Tsec'].values
        self.axis_t_samp = detecFunction['Tsamp'].values
        self.timeseries = detecFunction['Value'].values
        self.ylabel = ylabel
        self.detections = detections

    def plotDetectionFunction(self, unit='sec', newFig=True):
        if newFig:
            plt.figure()
        if unit == 'sec':
            xaxis = self.axis_t_sec
            xlabel = 'Time (sec)'
        elif unit == 'samp':
            xaxis = self.axis_t_samp
            xlabel = 'Time (sample)'
        else:
            raise ValueError('Invalid unit value. Should be "sec" or "samp"')
        # plot figure
        plt.plot(xaxis, self.timeseries, color='black')
        plt.xlabel(xlabel)
        plt.ylabel(self.ylabel)
        plt.axis([xaxis[0], xaxis[-1], min(self.timeseries), max(self.timeseries)])
        if len(self.detections) > 0:
            plt.plot(self.detections['startTimeSec'], self.detections['confidence'], '.r')
            plt.legend('Detections')
        plt.grid()
        plt.show()


@jit
def findPeaks(Kurt, threshold, delta_sec):
    # find peaks
    Pidx = spsig.argrelmax(Kurt['Value'].values)
    Pidx = Pidx[0][:]
    Pval = Kurt['Value'].values[Pidx]
    # only keep peaks >= threshold
    idx = Pval >= threshold
    Pidx = Pidx[idx]
    Pval = Pval[idx]
    Psec = Kurt.ix[Pidx]['Tsec'].values
    Psamp = Kurt.ix[Pidx]['Tsamp'].values
    Peaks = pd.DataFrame()
    Peaks['Value'] = []
    Peaks['Tsamp'] = []
    Peaks['Tsec'] = []
    while len(Pval) > 0:
        dt = Psec-Psec[0]
        idx = dt < delta_sec
        tmp = pd.DataFrame({'Value': [Pval[0]], 'Tsamp': [Psamp[0]], 'Tsec': [Psec[0]]})
        Peaks = pd.DataFrame.append(Peaks, tmp, ignore_index=True)
        for i in range(sum(idx)):
            Pidx = np.delete(Pidx, [0])
            Pval = np.delete(Pval, [0])
            Psec = np.delete(Psec, [0])
            Psamp = np.delete(Psamp, [0])
    return Peaks


# -*- coding: utf-8 -*-
"""
Created on Thu Feb 20 10:02:51 2020

@author: xavier.mouy
"""


class BaseClass(object):
    """
    Base class for all detectors.

    All detectors need to inheritate from this BaseClass in order to be built
    by the DetectorFactory.
    """

    def __init__(self, detector_name):
        self.detector_name = detector_name

    @classmethod
    def is_detector_for(cls, detector_name):
        """
        Check detector name.

        Compare the requested detector_name with each detector classe
        available.

        Parameters
        ----------
        detector_name : str
        Name of the detector class.

        Returns
        -------
        bool
        """
        return detector_name == cls.__name__


def DetectorFactory(detector_name, *args, **kwargs):
    """
    Detector Factory.

    Loads the detector class defined by detector_name. Each detector class must
    be added to the __init__.py file.

    For example :
    from .detector1 import Detector1

    where Detector1 is the name of the detector class to load from teh .py
    file detection1

    Parameters
    ----------
    detector_name : str
        Name of the detector class.
    *args : any
        Input arguments for the detector selected.
    **kwargs : any
        Keyword arguments for the detector selected.

    Raises
    ------
    ValueError
        If detector_name doesn't correspond to any of the detectors available.

    Returns
    -------
    Detector object

    """
    detec_list = []
    for cls in BaseClass.__subclasses__():
        detec_list.append(cls.__name__)
        if cls.is_detector_for(detector_name):
            return cls(detector_name, *args, **kwargs)
    raise ValueError("Invalid detector name. Detectors available: "
                     + str(detec_list))



# -*- coding: utf-8 -*-
"""
Created on Tue Jan  7 11:23:51 2020

@author: xavier.mouy
"""

import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import numpy as np
from scipy import signal, ndimage
from dask import delayed, compute
import copy
from numba import njit
import dask
import dask.array
import dask_image.ndfilters

## TODO: change Asserts by Raise

class Spectrogram:
    """A class for spectrograms.

    The Spectrogram object computes, denoises, and displays a spectrogram from
    a Sound object.

    Attributes
    ----------
    sampling_frequency : float
        Sampling frequency of the sound data.
    time_resolution : float
        Time resolution of the spectrogram, in seconds.
    frequency_resolution : float
        Frequency resolution of the spectrogram, in Hz.
    frame_samp : int
        Frame size, in samples
    frame_sec : float
        Frame size, in seconds
    step_samp : int
        Time step between conscutive spectrogram frames, in samples.
    step_sec : float
        Time step between conscutive spectrogram frames, in seconds.
    overlap_perc : float
        Percentage of overlap between conscutive spectrogram frames.
    overlap_samp : int
        Overlap between conscutive spectrogram frames, in samples.
    fft_samp : int
        Size to the Fast Fourier Transform, in samples.
    fft_sec : float
        Size to the Fast Fourier Transform, in seconds.
    window_type : str
        Type of weighting window applied to the signal before the FFT.
    axis_frequencies : numpy.ndarray
        1-D array with frequency values, in Hz, for each spectrogram rows.
    axis_times : numpy.ndarray
        1-D array with time values, in seconds, for each spectrogram column.
    spectrogram : numpy.ndarray
        2-D array with spectrogram energy data.

    Methods
    -------
    compute(sig, fs)
        Compute spectrogram.
    crop(frequency_min, frequency_max)
        Crop frequencies from the spectrogram.
    denoise(method, **kwargs)
        Denoise the spectrogram using various methods. 
        Methods implemented:
        METHODS           :    INPUT ARGUMENTS
        'median_equalizer':    window_duration in seconds.
    """

    _valid_units = ('samp', 'sec')
    _valid_windows = ('hann',)
    def __init__(self, frame, window_type, fft, step, sampling_frequency, unit='sec'):
        """
        Initialize Spectrogram object.

        Defines spectrogram parameters. 'step' and 'fft' sizes can be defined
        in seconds (unit='sec') or in samples (unit='samp'). The 'fft' size is
        automatically adjusted to the next power of two. Zero padding is
        possible by defining a 'fft' size greater than the 'frame' size.

        Parameters
        ----------
        frame : float
            Frame size in seconds or samples, depending on 'unit'.
        window_type : str
            Weighting window to teh signal before the FFT. Currently, only
            'hann' is supported.
        fft : float
            Size of the Fast Fourier Transform, in seconds or samples,
            depending on 'unit'.
        step : float
            Time step between conscutive spectrogram frames. In samples or 
            seconds depending on 'unit'.
        sampling_frequency : float
            Sampling frequency of the signal, in Hz.
        unit : str, optional
            Unit used when defining the 'frame' and 'fft' parameters. For
            seconds, use 'sec'. For samples, use 'samp'. The default is 'sec'.

        Returns
        -------
        None. Spectrogram object.

        """
        # Validation of the imput parameters
        assert (unit in Spectrogram._valid_units), ("Wrong unit value. Valid \
                                           units: ", Spectrogram._valid_units)
        assert fft >= frame, " fft should alwyas be >= frame"
        assert step < frame, "step should always be <= frame"
        assert (window_type in Spectrogram._valid_windows), ("Wrong window type\
                                . Valid values: ", Spectrogram._valid_windows)

        # Convert units in seconds/samples
        self._frame_samp, self._fft_samp, self._step_samp, self._frame_sec,\
        self._fft_sec, self._step_sec, self._overlap_perc, self._overlap_samp =\
        Spectrogram._convert_units(frame, fft, step, sampling_frequency, unit)

        # Time and frequency resolution
        self._sampling_frequency = sampling_frequency
        self._time_resolution = self.step_sec
        self._frequency_resolution = self.sampling_frequency / self.fft_samp

        # Define all other instance attributes
        self._window_type = window_type
        self._spectrogram = []
        self._axis_frequencies = []
        self._axis_times = []

    def _convert_units(frame, fft, step, sampling_frequency, unit):
        """Convert frame, fft, and step values to samples/seconds"""
        if unit == 'sec':
            frame_samp = round(frame*sampling_frequency)
            fft_samp = adjust_FFT_size(round(fft*sampling_frequency))
            step_samp = round(step*sampling_frequency)
            frame_sec = frame
            fft_sec = fft_samp*sampling_frequency
            step_sec = step
        elif unit == 'samp':
            frame_samp = frame
            fft_samp = adjust_FFT_size(fft)
            step_samp = step
            frame_sec = frame/sampling_frequency
            fft_sec = fft_samp/sampling_frequency
            step_sec = step/sampling_frequency
        overlap_samp = frame_samp-step_samp
        overlap_perc = (overlap_samp/frame_samp)*100
        return frame_samp, fft_samp, step_samp, frame_sec, fft_sec, step_sec, overlap_perc,overlap_samp

    def _compute_old(self, sig):
        """
        Compute spectrogram.

        Compute spectrogram from sound signal and return values in the object
        attribute 'spectrogram'.

        Parameters
        ----------
        sig : Sound object
            Sound object (core.audiotools.Sound) with the signal to anayse.

        Returns
        -------
        Populate the 'spectrogram' attribute of the Spectrogram object.
        axis_frequencies, numpy.ndarray
            1-D array with time axis values, in seconds.
        axis_times, numpy.ndarray
            1-D array with frequency axis values, in Hertz.
        spectrogram, numpy.ndarray
            2-D array with spectrogram values.

        """
        assert sig.waveform_sampling_frequency == self.sampling_frequency, "The sampling frequency provided doesn't match the one from the Spectrogram object."
        # Weighting window
        if self.window_type == 'hann':
            window = signal.hann(self.frame_samp)
        # Calculates  spectrogram
        self._axis_frequencies, self._axis_times, self._spectrogram = signal.spectrogram(sig.waveform, fs=self.sampling_frequency, window=window, noverlap=self.overlap_samp, nfft=self.fft_samp, scaling='spectrum')
        self._spectrogram = 20*np.log10(self._spectrogram)
        return self._axis_frequencies, self._axis_times, self._spectrogram

    def compute(self, sig, dB=False, use_dask=False, dask_chunks=40):
        """
        Compute spectrogram.

        Compute spectrogram from sound signal and return values in the object
        attribute 'spectrogram'.

        Parameters
        ----------
        sig : Sound object
            Sound object (core.audiotools.Sound) with the signal to anayse.
        dB : bool, optional
            If set to True, returns spectrogram values in dB. The default is
            False.
        use_dask : bool, optional
            If set to True, parallelize the computation using Dask. The default
            is False.
        dask_chunks : int, optional
            Set the number of chunks the signal is broken in to for the
            parallelized computation using Dask. Only used if use_dask is set
            to True. The default is 40.

        Returns
        -------
        Populate the 'spectrogram' attribute of the Spectrogram object.
        axis_frequencies, numpy.ndarray
            1-D array with time axis values, in seconds.
        axis_times, numpy.ndarray
            1-D array with frequency axis values, in Hertz.
        spectrogram, numpy.ndarray
            2-D array with spectrogram values.
        """
        # Weighting window
        if self.window_type == 'hann':
            win = np.hanning(self.frame_samp)
        #sig = self.waveform
        #step = self.frame_samp - self.overlap_samp
        starts = np.arange(0,len(sig.waveform)-self.frame_samp, self.step_samp, dtype=int)
        stops = starts + self.frame_samp
        start_chunks = np.array_split(starts, dask_chunks)
        stop_chunks = np.array_split(stops, dask_chunks)
        spectrogram = []
        idx=0
        for start_chunk, stop_chunk in zip(start_chunks, stop_chunks):
            sig_chunk = sig.waveform[start_chunk[0]:stop_chunk[-1]]
            chunk_size = len(start_chunk)
            if use_dask:
                spectro_chunk = delayed(Spectrogram._calc_spectrogram)(sig_chunk,win,
                                           start_chunk-start_chunk[0],
                                           stop_chunk-start_chunk[0],
                                           self.fft_samp)
            else:
                spectro_chunk = Spectrogram._calc_spectrogram(sig_chunk,win,
                                start_chunk-start_chunk[0],
                                stop_chunk-start_chunk[0],
                                self.fft_samp)
            spectrogram.append(spectro_chunk)
            idx += chunk_size
        if use_dask:
            spectrogram = compute(spectrogram)
            spectrogram = np.concatenate(spectrogram[0][:], axis=1)
        else:
            spectrogram = np.concatenate(spectrogram[:], axis=1)
        if dB:
            spectrogram = 20*np.log10(spectrogram)
        self._spectrogram = spectrogram
        self._axis_times = starts/self.sampling_frequency  # ?? needed ?
        self._axis_frequencies = np.arange(0,self._sampling_frequency/2,self._frequency_resolution) # ?? needed ?
        return self._axis_frequencies, self._axis_times, self._spectrogram

    @staticmethod
    #@njit
    def _calc_spectrogram(sig, win, starts, stops, fft_samp):
        fft_samp = np.int32(fft_samp)
        fnyq0 = np.round(fft_samp/2)
        fnyq = np.int(fnyq0)
        spectro = np.empty((fnyq, len(starts)))  # the default
        idx = 0
        for start, stop in zip(starts, stops):
            s = sig[start:stop]*win
            Spectrum = np.fft.fft(s, fft_samp)
            Spectrum = abs(Spectrum)  # amplitude
            Spectrum = Spectrum*2
            # Spectrum = Spectrum**2
            # ts_window = getFFT(sig[start:stop],fft_samp)
            spectro[:, idx] = Spectrum[0:fnyq]
            idx += 1
        return spectro

    def crop(self, frequency_min=None, frequency_max=None, time_min=None, time_max=None, inplace=False):
        """
        Crop frequencies from the spectrogram.

        Crop the spectrogram matrix by keeping only frequency rows above
        frequency_min and below frequency_max. If frequency_min is not provided
        then, only spectrogram rows above frequency_max will be removed. If
        frequency_max is not provided then, only spectrogram rows below
        frequency_min will be removed. The axis_frequencies attribute of the
        spectrogram object are automatically updated.

        Parameters
        ----------
        frequency_min : float, optional
            Minimum frequency limit, in Hz. The default is None.
        frequency_max : float, optional
            Maximum frequency limit, in Hz. The default is None.
        time_min : float, optional
            Minimum time limit, in sec. The default is None.
        time_max : float, optional
            Maximum time limit, in sec. The default is None.
        inplace : bool, optional
            If True, do operation inplace and return None. The default is False

        Returns
        -------
        None. Cropped spectrogram matrix.

        """
        # Find frequency indices
        if frequency_min is None:
            min_row_idx = 0
        else:
            min_row_idx = np.where(self._axis_frequencies < frequency_min)
            if np.size(min_row_idx) == 0:
                min_row_idx = 0
            else:
                min_row_idx = min_row_idx[0][-1] + 1
        if frequency_max is None:
            max_row_idx = self._axis_frequencies.size-1
        else:
            max_row_idx = np.where(self._axis_frequencies > frequency_max)
            if np.size(max_row_idx) == 0:
                max_row_idx = self._axis_frequencies.size-1
            else:
                max_row_idx = max_row_idx[0][0]
        # Find time indices
        if time_min is None:
            min_col_idx = 0
        else:
            min_col_idx = np.where(self._axis_times < time_min)
            if np.size(min_col_idx) == 0:
                min_col_idx = 0
            else:
                min_col_idx = min_col_idx[0][-1] + 1
        if time_max is None:
            max_col_idx = self._axis_times.size-1
        else:
            max_col_idx = np.where(self._axis_times > time_max)
            if np.size(max_col_idx) == 0:
                max_col_idx = self._axis_times.size-1
            else:
                max_col_idx = max_col_idx[0][0]
        # update spectrogram and axes
        if inplace:
            # self._axis_frequencies = self._axis_frequencies[min_row_idx:max_row_idx]
            # self._axis_times = np.arange(0,(max_col_idx - min_col_idx)*self._time_resolution,self._time_resolution)
            # self._spectrogram = self._spectrogram[min_row_idx:max_row_idx, min_col_idx:max_col_idx]   
            # out_object = None
            self._axis_frequencies = self._axis_frequencies[min_row_idx:max_row_idx+1]
            self._axis_times = self._axis_times[min_col_idx:max_col_idx+1]-self._axis_times[min_col_idx]
            self._spectrogram = self._spectrogram[min_row_idx:max_row_idx+1, min_col_idx:max_col_idx+1]   
            out_object = None
        else:
            out_object = copy.copy(self)
            out_object._axis_frequencies = out_object._axis_frequencies[min_row_idx:max_row_idx+1]
            #out_object._axis_times = np.arange(0,(max_col_idx - min_col_idx)*out_object._time_resolution,out_object._time_resolution)
            out_object._axis_times = out_object._axis_times[min_col_idx:max_col_idx+1]-out_object._axis_times[min_col_idx]
            #out_object._spectrogram = out_object._spectrogram[min_row_idx:max_row_idx, min_col_idx:max_col_idx]   
            out_object._spectrogram = out_object._spectrogram[min_row_idx:max_row_idx+1, min_col_idx:max_col_idx+1]
            if out_object._spectrogram.shape[1] != len(out_object._axis_times):
                raise ValueError("Spectrogram axes don't match spectrogram matrix.")
        return out_object

    def denoise(self, method, **kwargs):
        """
        Denoise spectrogram.

        Denoise the spectrogram using various methods. The methods implemented
        are:
            METHODS           :    INPUT ARGUMENTS
            'median_equalizer':    window_duration in seconds.
                                   inplace 

        Parameters
        ----------
        method : str
            DESCRIPTION.
        **kwargs : variable
            Parameters for the methods selected.

        Raises
        ------
        ValueError
            If method is not valid.

        Returns
        -------
        None. Denoised spectrogram matrix.

        """
        denoise_methods = ('median_equalizer',)
        if method in denoise_methods:
            eval("self._" + method + "(**kwargs)")
        else:
            raise ValueError('Method not recognized. Methods available:'
                             + str(denoise_methods))

    def _median_equalizer(self, window_duration, use_dask=False, dask_chunks=(1000,1000), inplace=False):
        """
        Median equalizer.

        Denoises the spectrogram matrix by subtracting the meidan spectrogram
        compuetd with a median filter of window "window_sixe" to the original
        spectrogram. Negative values of teh denoised spectrogram are set to
        zero.

        Parameters
        ----------
        window_duration : float
            Durations of the median filter, in seconds.
        use_dask : bool, optional
            If True, to parallelize the processing with Dask. The default is
            False
        dask_chunks : tuple-> (int, int), optional
            Size of the sepctrogram chunks for the parallelized computing in
            number of bins. dask_chunks=(number of frequency bins, number of
            time bins). Only used if use_dask is True. The default is
            (1000,1000).
        inplace : bool, optional
            If True, do operation inplace and return None. The default is False

        Returns
        -------
        Denoised spectrogram matrix.

        """
        if use_dask:
            dask_spectro = dask.array.from_array(self._spectrogram, chunks=dask_chunks)
            Smed = dask_image.ndfilters.median_filter(dask_spectro,
                                                       size=(1,round(window_duration/self.time_resolution)),
                                                       mode='mirror')
            Smed = Smed.compute()
        else:
            Smed = ndimage.median_filter(self._spectrogram, (1,round(window_duration/self.time_resolution)))
        if inplace:
            self._spectrogram = self._spectrogram-Smed
            self._spectrogram[self._spectrogram < 0] = 0  # floor
            out_object = None
        else:
            out_object = copy.copy(self)
            out_object._spectrogram = out_object._spectrogram-Smed
            out_object._spectrogram[out_object._spectrogram < 0] = 0  # floor            
        return out_object

    @property
    def frame_samp(self):
        """Return the frame_samp attribute."""
        return self._frame_samp

    @property
    def frame_sec(self):
        """Return the frame_sec attribute."""
        return self._frame_sec

    @property
    def step_samp(self):
        """Return the step_samp attribute."""
        return self._step_samp

    @property
    def step_sec(self):
        """Return the step_sec attribute."""
        return self._step_sec

    @property
    def fft_samp(self):
        """Return the fft_samp attribute."""
        return self._fft_samp

    @property
    def fft_sec(self):
        """Return the fft_sec attribute."""
        return self._fft_sec

    @property
    def overlap_perc(self):
        """Return the overlap_perc attribute."""
        return self._overlap_perc

    @property
    def overlap_samp(self):
        """Return the overlap_samp attribute."""
        return self._overlap_samp

    @property
    def sampling_frequency(self):
        """Return the sampling_frequency attribute."""
        return self._sampling_frequency

    @property
    def time_resolution(self):
        """Return the time_resolution attribute."""
        return self._time_resolution

    @property
    def frequency_resolution(self):
        """Return the frequency_resolution attribute."""
        return self._frequency_resolution

    @property
    def window_type(self):
        """Return the window_type attribute."""
        return self._window_type

    @property
    def axis_frequencies(self):
        """Return the axis_frequencies attribute."""
        return self._axis_frequencies

    @property
    def axis_times(self):
        """Return the axis_times attribute."""
        return self._axis_times

    @property
    def spectrogram(self):
        """Return the spectrogram attribute."""
        return self._spectrogram

def adjust_FFT_size(nfft):
        """ Adjust nfft to the next power of two if necessary."""
        nfft_adjusted = next_power_of_2(nfft)
        if nfft_adjusted != nfft:
            print('Warning: FFT size automatically adjusted to', nfft_adjusted, 'samples (original size:', nfft,')')
        return nfft_adjusted

def next_power_of_2(x):
    """Calculate the next power of two for x."""
    return 1 if x == 0 else 2**(x - 1).bit_length()



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 11:52:11 2020

@author: xavier.mouy
"""
#import pytest
import os
from ecosound.core.annotation import Annotation


def get_paths():
    """ Define paths of the test data."""
    paths = {'test_dir': os.path.dirname(os.path.realpath(__file__))}
    paths['data_dir'] = os.path.abspath(os.path.join(os.path.dirname( __file__ ), os.pardir, 'data'))
    paths['raven_annot_dir'] = os.path.join(paths['data_dir'],'Raven_annotations')
    paths['pamlab_annot_dir'] = os.path.join(paths['data_dir'],'PAMlab_annotations')
    paths['wav_files_dir'] = os.path.join(paths['data_dir'],'wav_files')

    # Raven annotations files
    raven_annot_files = []
    raven_annot_files.append({'filename': '67674121.181018013806.Table.1.selections.txt',
                              'annot_number': 773,
                              'duplicates': 0})
    raven_annot_files.append({'filename': 'AMAR173.4.20190916T061248Z.Table.1.selections.txt',
                              'annot_number': 1114,
                              'duplicates': 557})
    paths['raven_annot_files'] = raven_annot_files
    return paths


def test_len_is_0():
    """ Test len of annot is 0 upon instantiation. """
    annot = Annotation()
    assert len(annot) == 0
    return None


def test_from_raven_singlefile(fileidx=0):
    """ Test number of annotations when importing a single Raven file.

    Only use one file for this test defined by fileidx=0
    """
    paths = get_paths()
    annot = Annotation()
    annot.from_raven(os.path.join(paths['raven_annot_dir'],paths['raven_annot_files'][fileidx]['filename']), verbose=False)
    assert len(annot) == paths['raven_annot_files'][fileidx]['annot_number'] - paths['raven_annot_files'][fileidx]['duplicates']
    return None

def test_from_raven_singlefile_with_duplicates(fileidx=1):
    """ Test that Raven annotation duplicates are removed correctly. """
    paths = get_paths()
    annot = Annotation()
    annot.from_raven(os.path.join(paths['raven_annot_dir'], paths['raven_annot_files'][fileidx]['filename']), verbose=False)
    assert len(annot) == paths['raven_annot_files'][fileidx]['annot_number'] - paths['raven_annot_files'][fileidx]['duplicates']
    return None


def test_from_raven_dir():
    """ Test number of annotations when importing full folder (Raven). """
    paths = get_paths()
    total_annotations = 0
    for file in paths['raven_annot_files']:
        total_annotations += file['annot_number'] - file['duplicates']
    annot = Annotation()
    annot.from_raven(paths['raven_annot_dir'], verbose=False)
    assert len(annot) == total_annotations
    return None

# print(len(annot2))





#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

exec(open("ecosound/_version.py").read())

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

# automatically captured required modules for install_requires in requirements.txt
with open('requirements.txt', encoding='utf-8') as f:
    requirements = f.read().split('\n')

setup_requirements = [ ]

test_requirements = [ ]

setup(
    name="ecosound", # Replace with your own username
    version=__version__,
	author="Xavier Mouy",
    author_email="xaviermouy@uvic.ca",
    description="Python toolkit for analysing passive acoustic data",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/x-rst",
    keywords='ecosound',
	url="https://github.com/xaviermouy/ecosound",    
    include_package_data=True,
	#package_data={'': ['core/*.json']},
    packages=find_packages(include=['ecosound', 'ecosound.*'],exclude=['docs', 'tests','resources']),
    install_requires=requirements,
	setup_requires=setup_requirements,
	test_suite='tests',
	tests_require=test_requirements,
    license="BSD license",
	classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License", 
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    python_requires='>=3.7.0,<3.8.0',
	zip_safe=False,
)


# -*- coding: utf-8 -*-
"""
This routine take takes false fish detections from the detector and turns them
as noise annotations for retraining the classifier

Created on Thu Nov  5 09:32:38 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement
import pandas as pd
import os

xls_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_manual_analysis\Noise_files_RCA_Out_Oct17_Dec3_2018_67391492.xlsx'
in_dir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_Out_Oct17_Dec3_2018_67391492'
out_dir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Noise_dataset\Noise_files_RCA_Out_Oct17_Dec3_2018_67391492'
fish_label = 'FS'
min_threshold = 0.7
noise_label = 'NN'

# load names of file and start/stop times where false alarms have been manually
# identified
df = pd.read_excel(xls_file,header=None)

for idx in range(0,len(df)):
    # file name to load
    wav_file_name = df[0][idx]
    tmin_sec = df[1][idx]
    tmax_sec = df[2][idx]
    print(wav_file_name, tmin_sec, tmax_sec)
    detec_file_path = os.path.join(in_dir, wav_file_name + '.nc')
    # load detection/measurement file
    meas = Measurement()
    meas.from_netcdf(detec_file_path)
    data_df = meas.data
    # Only keep fish detections above the given confidence threshold and times
    data_df_filt = data_df[(data_df.label_class==fish_label)
                           & (data_df.confidence>=min_threshold)
                           & (data_df.time_min_offset>=tmin_sec)
                           & (data_df.time_max_offset<=tmax_sec)
                           ]
    data_df_filt.reset_index(inplace=True, drop=True)
    meas.data = data_df_filt
    # Change fish labels to noise labels
    meas.insert_values(label_class=noise_label)
    # Save to new nc file
    meas.to_netcdf(os.path.join(out_dir,wav_file_name + str(idx)))

print('done')


# -*- coding: utf-8 -*-
"""
Created on Sun Mar  1 21:26:33 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..") # Adds higher directory to python modules path.
from ecosound.core.audiotools import Sound
from scipy import signal
import matplotlib.pyplot as plt
import numpy as np
import time
from numpy.fft import fft
from dask import delayed, compute
# from dask.distributed import Client, LocalCluster
# cluster = LocalCluster()
# client = Client(cluster,processes=False)


single_channel_file = r"../ecosound/resources/67674121.181018013806.wav"


# start and stop time of wavfile to analyze
t1 = 1
t2 = 120#1000
## ###########################################################################


# load audio data
sound = Sound(single_channel_file)
sound.read(channel=0, chunk=[t1, t2], unit='sec')

# Calculates  spectrogram
frame_samp = 3000
overlap_samp = 2500
fft_samp = 4096

# Using scipy #########################################
def spectro_numpy(frame_samp,overlap_samp,fft_samp):
    tic = time.perf_counter()
    axis_frequencies, axis_times, spectrogram = signal.spectrogram(sound.waveform,
                                                                    fs=sound.waveform_sampling_frequency,
                                                                    window=signal.hann(frame_samp),
                                                                    noverlap=overlap_samp,
                                                                    nfft=fft_samp,
                                                                    scaling='spectrum')
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")
#fig, ax = plt.subplots(figsize=(16,4), sharex=True)
#im = ax.pcolormesh(axis_times,axis_frequencies,spectrogram, cmap='jet',vmin = np.percentile(spectrogram,50), vmax= np.percentile(spectrogram,99.9))

def getFFT(sig,nfft):
    return abs(np.fft.fft(sig,nfft))

def crop(x,fft_samp):
    return x[0:int(fft_samp/2)]

def spectro_loop(frame_samp,overlap_samp,fft_samp):
    tic = time.perf_counter()
    starts = np.arange(0,len(sound.waveform),frame_samp-overlap_samp,dtype=int)
    starts = starts[starts + frame_samp < len(sound.waveform)]
    xns = []
    for start in starts:
        # short term discrete fourier transform
        ts_window = getFFT(sound.waveform[start:start + frame_samp],fft_samp)
        ts_window2 = crop(ts_window,fft_samp)
        xns.append(ts_window2)
    specX = np.array(xns).T
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")

def slice(sound, start, frame_samp):
    return sound.waveform[start:start + frame_samp]

def spectro_loop_dask(frame_samp,overlap_samp,fft_samp):
    tic = time.perf_counter()
    #client = Client(n_workers=4)
    starts = np.arange(0,len(sound.waveform),frame_samp-overlap_samp,dtype=int)
    starts = starts[starts + frame_samp < len(sound.waveform)]
    xns = []
    for start in starts:
        # short term discrete fourier transform
        sig = delayed(slice)(sound,start,frame_samp)
        ts_window1 = delayed(fft)(sig,fft_samp)
        ts_window2 = delayed(abs)(ts_window1)
        #abs(
        #ts_window2 = delayed(crop)(ts_window,fft_samp)
        xns.append(ts_window2)
    #specX = delayed(np.array(xns).T)
    #specX = delayed(abs)(xns)
    #visualize(xns)
    compute(xns)
    #specX = specX.compute()
    toc = time.perf_counter()
    print(f"Dexecuted in {toc - tic:0.4f} seconds")
    #client.close()
    
# ## Using Dask #########################################
# from dask import delayed
# from dask.distributed import Client, progress
# client = Client(n_workers=4)


# def getFFT(sig):
#     return abs(np.fft.fft(sig))

# def crop(x,fft_samp):
#     return x[0:int(fft_samp/2)]

# tic = time.perf_counter()
# starts = np.arange(0,len(sound.waveform),fft_samp-overlap_samp,dtype=int)
# starts = starts[starts + fft_samp < len(sound.waveform)]
# xns = []
# for start in starts:
#     # short term discrete fourier transform
#     ts_window = delayed(getFFT)(sound.waveform[start:start + fft_samp])
#     ts_window2 = delayed(crop)(ts_window,fft_samp)
#     xns.append(ts_window2)

# specX = delayed(np.array(xns).T)
# specX.visualize()
# # #spec = delayed(10*np.log10(specX))
# # result = specX.compute()

# # toc = time.perf_counter()
# # print(f"Dexecuted in {toc - tic:0.4f} seconds")

# #fig2, ax2 = plt.subplots(figsize=(16,4), sharex=True)
# #im = ax2.pcolormesh(spec, cmap='jet',vmin = np.percentile(spec,50), vmax= np.percentile(spec,99.9))

# import time
# from dask.distributed import Client
# from dask import delayed
# from time import sleep
    
# ## Using Dask #########################################

# def inc(x):
#         sleep(1)
#         print('inc')
#         return x + 1

# def add(x, y):
#     sleep(1)
#     print('add')
#     return x + y
    
# def test():

#     client = Client(n_workers=10)
    
#     tic = time.perf_counter()
#     x = delayed(inc)(1)
#     y = delayed(inc)(2)
#     z = delayed(add)(x, y)
#     total = z.compute()
#     print(total)
#     toc = time.perf_counter()
#     print(f"Downloaded the tutorial in {toc - tic:0.4f} seconds")
#     client.close()

if __name__ == '__main__':
    spectro_numpy(frame_samp,overlap_samp,fft_samp)
    spectro_loop(frame_samp,overlap_samp,fft_samp)
    spectro_loop_dask(frame_samp,overlap_samp,fft_samp)



# -*- coding: utf-8 -*-
"""
Created on Thu Jan 30 11:30:08 2020

@author: xavier.mouy
"""
import functools
import time


def listinput(func):
    """Set input argument of function as a list if not already the case."""
    @functools.wraps(func)
    def wrapper_listinput(*args, **kwargs):
        #  print(type(**kwargs))
        if type(*args) is not list:
            value = func([*args], **kwargs)
        else:
            value = func(*args, **kwargs)
        return value
    return wrapper_listinput


def timeit(func):
    """Print time it took to exceute function."""
    @functools.wraps(func)
    def wrapper_listinput(*args, **kwargs):
        tic = time.perf_counter()
        value = func(*args, **kwargs)
        toc = time.perf_counter()
        print(f"Executed in {toc - tic:0.4f} seconds")
        return value
    return wrapper_listinput



# -*- coding: utf-8 -*-
"""
Created on Fri May 22 14:04:01 2020

@author: xavier.mouy
"""


# import sys
# sys.path.append("..") # Adds higher directory to python modules path.
# from ecosound.core.audiotools import Sound
# from ecosound.core.spectrogram import Spectrogram
# from ecosound.core.annotation import Annotation
# from ecosound.core.measurement import Measurement
# from ecosound.visualization.grapher_builder import GrapherFactory
# from ecosound.measurements.measurer_builder import MeasurerFactory
# import time


# from ecosound.core.measurement import Measurement
# # load detections
# detection_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\results\67674121.181017060806.wav.nc"
# detec = Measurement()
# detec.from_netcdf(detection_file)
# detec.summary()

# from ecosound.core.measurement import Measurement
# from ecosound.core.audiotools import Sound
# from ecosound.core.spectrogram import Spectrogram
# from ecosound.visualization.grapher_builder import GrapherFactory

# # load detections
# detection_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\results\67674121.181017060806.wav.nc"
# detec = Measurement()
# detec.from_netcdf(detection_file)

# # load audio file (first 15 sec only)
# audio_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\data\67674121.181017060806.wav"
# sound = Sound(audio_file)
# sound.read(channel=0, chunk=[0, 15], unit='sec', detrend=True)

# # Calculate  spectrogram
# frame = 3000
# nfft = 4096
# step = 500
# window_type = 'hann'
# spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='samp')
# spectro.compute(sound, dB=True)
# #spectro.crop(frequency_min=0, frequency_max=1000, inplace=True)

# # Plot
# graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000)
# graph.add_data(sound) # add waveform data
# graph.add_data(spectro) # add spectrogram
# graph.add_annotation(detec, panel=0, color='green', label='Detections') # overlay detections on waveform plot
# graph.add_annotation(detec, panel=1, color='green', label='Detections') # overlay detections on spectrogram plot
# graph.colormap = 'binary'
# #graph.colormap = 'jet'
# graph.show()





from ecosound.core.measurement import Measurement
import seaborn as sns
# load detections
detection_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\results\67674121.181017060806.wav.nc"
detec = Measurement()
detec.from_netcdf(detection_file)

detec.data['freq_peak'].groupby(detec.data['label_class'])
fig = sns.violinplot(y=detec.data['label_class'], x=detec.data['freq_peak'])
fig.set(xlabel='Detections peak frequency (Hz)', ylabel='Detections class')
#sns.show()

# ## Input paraneters ##########################################################


# audio_file = r"C:\Users\xavier.mouy\Desktop\Tutorial\data\67674121.181017060806.wav"

# #audio_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\audio_data\AMAR173.4.20190916T011248Z.wav"
# #annotation_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\datasets\UVIC_hornby-island_2019\manual_annotations\AMAR173.4.20190916T011248Z.Table.1.selections.txt"
# #detection_file = r"C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Full_dataset\AMAR173.4.20190916T011248Z.wav.nc"


# # Spectrogram parameters
# frame = 3000
# nfft = 4096
# step = 500
# #ovlp = 2500
# fmin = 0
# fmax = 1000
# window_type = 'hann'

# # start and stop time of wavfile to analyze
# t1 = 00#24
# t2 = 60#40
# ## ###########################################################################
# tic = time.perf_counter()

# # load audio data
# sound = Sound(audio_file)
# sound.read(channel=0, chunk=[t1, t2], unit='sec', detrend=True)

# # Calculates  spectrogram
# spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='samp')
# spectro.compute(sound, dB=True, use_dask=True, dask_chunks=40)
# spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)

# # load annotations
# annot = Annotation()
# annot.from_raven(annotation_file)

# # load detections
# detec = Measurement()
# detec.from_netcdf(detection_file)

# toc = time.perf_counter()
# print(f"Executed in {toc - tic:0.4f} seconds")

# # Plot
# graph = GrapherFactory('SoundPlotter', title='Recording', frequency_max=1000, time_min=30)
# graph.add_data(sound)
# graph.add_annotation(detec, panel=0, color='green', label='Detections')
# graph.add_data(spectro)
# graph.add_annotation(annot, panel=1, color='red', label='Annotations')
# graph.add_annotation(detec, panel=1, color='green', label='Detections')
# graph.colormap = 'binary'
# #graph.colormap = 'jet'
# graph.show()


# -*- coding: utf-8 -*-
"""
Created on Thu Nov 12 14:21:12 2020

@author: xavier.mouy
"""
import sys
sys.path.append("..")  # Adds higher directory to python modules path.
#from ecosound.measurements.measurer_builder import MeasurerFactory
import pickle
import numpy as np

class Classifier():
    
    def __init__(self):
        self.model = None
        self.features = None
        self.Norm_mean = None
        self.Norm_std = None
        self.classes_encoder = None        
        
    def load_model(self, model_file):
        # laod classif model      
        classif_model = pickle.load(open(model_file, 'rb'))  
        self.features = classif_model['features']
        self.model = classif_model['model']
        self.Norm_mean = classif_model['normalization_mean']
        self.Norm_std = classif_model['normalization_std']
        self.classes_encoder = classif_model['classes']
        
    def classify(self, measurements):
        # data dataframe
        data = measurements.data
        n1=len(data)
        # drop observations/rows with NaNs
        data = data.replace([np.inf, -np.inf], np.nan)
        data.dropna(subset=self.features, axis=0, how='any', thresh=None, inplace=True)
        n2=len(data)
        # Classification - predictions
        X = data[self.features]
        X = (X-self.Norm_mean)/self.Norm_std
        pred_class = self.model.predict(X)
        pred_prob = self.model.predict_proba(X)
        pred_prob = pred_prob[range(0,len(pred_class)),pred_class]
        # Relabel
        for index, row in self.classes_encoder.iterrows():
            pred_class = [row['label'] if i==row['ID'] else i for i in pred_class]
        # update measurements
        data['label_class'] = pred_class
        data['confidence'] = pred_prob
        measurements.data = data
        return measurements
            


# -*- coding: utf-8 -*-
"""
Created on Fri May 19 15:26:24 2017

@author: xavier.mouy
"""
# --------------------------------------------------------------
##TODO: resample waveform
##TODO: play sound
# --------------------------------------------------------------

import soundfile as sf
import os
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as spsig
import copy
import ecosound.core.tools


class Sound:
    """
    A class to load and manipulate a sound file

    This class can load data from an entire, or part of a, sound file, filter
    the loaded data, select subsections, and plotteh waveform. Currently a
    Sound object can only load data from one channel at a time.

    Attributes
    ----------
    file_full_path : str
        Path of the sound file, including file name and extension.
    file_dir : str
        Path of the sound file directory.
    file_name : str
        Name of the sound file withjout teh extension.
    file_extension : str
        Extension of the sound file (e.g. ".wav").
    file_sampling_frequency : int
        Sampling frequency of the original sound data, in Hz.
    file_duration_sample : float
        Duration of the sound data from the file, in samples.
    file_duration_sec : float
        Duration of the sound data from the file, in seconds.
    channels : int
        Number of channels available in the sound file.
    channel_selected : int
        Channel from which the waveform data was loaded from.
    waveform : numpy.ndarray
        Waveform of the loaded data for the selected channel (channel_selected)
        and time frame selected.
    waveform_sampling_frequency : float
        Sampling frequency of the loaded waveform data. It can differ from
        file_sampling_frequency if the waveform was up- or down- sampled.
    waveform_start_sample : float
        Index of the first sample of the loaded waveform data relative to the
        begining of the sound file.
    waveform_stop_sample : float
        Index of the last sample of the loaded waveform data relative to the
        begining of the sound file.
    waveform_duration_sample : float
        Duration of the loaded waveform data, in samples.
    waveform_duration_sec : float
        Duration of the loaded waveform data, in seconds.
    filter_applied : bool
        True if the waveform data was filtered.
    filter_parameters : Filter obj
        Filter object with all filter paramters and coefficients. Empty if no
        filter was applied.

    Methods
    -------
    read(channel=0, chunk=[])
        Reads a sound file with the option to select a specific channel and
        read only a section of the file.
    filter(filter_type, cutoff_frequencies, order=4)
        Applies a scientific filter on the audio signal
    plot_waveform(unit='sec', newfig=False, title='')
        Displays a graph with the waveform of the audio signal
    select_snippet(chunk)
        Extract a chunk of the waveform as a new Sound object
    tighten_waveform_window(energy_percentage)
        Crops the beginning  and end times of a waveform in a Sound object
        based on a percentage of energy.

    """

    def __init__(self, infile):
        """
        Initialize Sound object.

        Parameters
        ----------
        infile : str
            Path of the sound file.

        Raises
        ------
        ValueError
            If sound file can't be found.

        Returns
        -------
        Sound object.

        """
        if os.path.isfile(infile):
            myfile = sf.SoundFile(infile)
            self._file_duration_sample = myfile.seek(0, sf.SEEK_END)
            self._file_sampling_frequency = myfile.samplerate
            self._file_duration_sec = self._file_duration_sample / \
                self._file_sampling_frequency
            self._channels = myfile.channels
            self._channel_selected = []
            self._file_dir = os.path.dirname(infile)
            self._file_name = os.path.basename(os.path.splitext(infile)[0])
            self._file_extension = os.path.splitext(infile)[1]
            self._filter_applied = False
            self._waveform = []
            self._waveform_start_sample = []
            self._waveform_stop_sample = []
            self._waveform_duration_sample = 0
            self._waveform_duration_sec = 0
            self._waveform_sampling_frequency = self._file_sampling_frequency
            self.detrended = []
            myfile.close()
        else:
            raise ValueError("The sound file can't be found. Please verify"
                             + ' sound file name and path')

    def read(self, channel=0, chunk=[], unit='samp', detrend=False):
        """
        Load data from sound file.

        Load data from a sound file with the option to select a specific
        channel and load only a section of the file. Data are loaded as a numpy
        arrayin in the object attribute "waveform".

        Parameters
        ----------
        channel : int, optional
            ID of the audio channel to load. The default is 0.
        chunk : list, optional
            List with two floats indicating the [start time, stop time], in
            samples, of the chunk of audio data to load. An empty list []
            loads data from the entire audio file. The default is [].
        unit : str, optional
            Time unit of the 'chunk' parameter. Can be set to 'sec' for seconds
            or 'samp', for samples. The default is 'samp'.
        detrend : bool, optional
            Remove DC offset of the waveform by subtracting the mean. The
            default is False.

        Raises
        ------
        ValueError
            If the chunk list has only 1 value.
            If the first value in the chunk list is greater or equal to the 
               second one.
            If values in the chunk list exceed the audio file limits.
            If the channel selected does not exist.
            If samp is not set to 'samp' or 'sec'

        Returns
        -------
        None. Load audio data in the waveform attribute and update all waveform
        related attributes.

        """
        # check that the channel id is valid
        if (channel >= 0) & (channel <= self._channels - 1):
            if len(chunk) == 0:  # read the entire file
                sig, fs = sf.read(self.file_full_path, always_2d=True)
                self._waveform = sig[:, channel]
                self._waveform_start_sample = 0
                self._waveform_stop_sample = self.file_duration_sample-1
                self._waveform_duration_sample = len(self._waveform)
                self._waveform_duration_sec = self._waveform_duration_sample/fs
            else:
                if unit not in ('samp','sec'):
                    raise ValueError('Invalid unit. Should be set to "sec" or'
                                     + '"samp".')
                # convert chunk to sampels if needed
                if unit in ('sec'):
                    chunk = np.round(np.dot(chunk,self.waveform_sampling_frequency))
                if len(chunk) == 2:  # only read a section of the file
                    # Validate input values
                    if (chunk[0] < 0) | (chunk[0] >= self.file_duration_sample):
                        raise ValueError('Invalid chunk start value. The sample'
                                          + ' value chunk[0] is outside of the'
                                          + ' file limits.')
                    elif (chunk[1] < 0) | (chunk[1] > self.file_duration_sample):
                        raise ValueError('Invalid chunk stop value. The sample'
                                         + ' value chunk[1] is outside of the'
                                         + ' file limits.')
                    elif chunk[1] <= chunk[0]:
                        raise ValueError('Invalid chunk values. chunk[1] must'
                                         + ' be greater than chunk[0]')
                    # read data
                    sig, fs = sf.read(self.file_full_path, start=int(chunk[0]),
                                      stop=int(chunk[1]), always_2d=True)
                    self._waveform = sig[:, channel]
                    self._waveform_start_sample = chunk[0]
                    self._waveform_stop_sample = chunk[1]
                    self._waveform_duration_sample = len(self._waveform)
                    self._waveform_duration_sec = self._waveform_duration_sample/fs
                else:
                    raise ValueError('Invalid chunk values. The argument chunk'
                                     + ' must be a list of 2 elements.')
            self._channel_selected = channel
            if detrend: # removes DC offset
                self._waveform = self._waveform - np.mean(self._waveform)

        else:
            msg = ''.join(['Channel ', str(channel), ' does not exist (',
                           str(self._channels), ' channels available).'])
            raise ValueError(msg)

    def filter(self, filter_type, cutoff_frequencies, order=4):
        """
        Filter the audio signal.

        Applies low-pass, high-pass, or band-pass scientific filter to the
        audio signal. The attribute waveform is updated with the filtered 
        signal. The same data can only be filtered once.

        Parameters
        ----------
        filter_type : str
            Type of filter. Can be set to 'bandpass', 'lowpass' or 'highpass'.            
        cutoff_frequencies : list
            Cutoff frequencies of the filter, in Hz (float). Must be a list with a
            single float if the filter_type is set to 'lowpass' or 'highpass'.
            Must be a list with two float values (i.e.,[fmin, fmax]) if the
            filter_type is set to 'bandpass'.
        order : int, optional
            Order of the filter. The default is 4.

        Raises
        ------
        ValueError
            If signal is filtered more than once.
            If the waveform attribute is empty
            If the filter type is not set to 'bandpass', 'lowpass', or 'highpass'
            If the cutoff_frequencies has not enough, or too much values for
            the filter type selected.
            If the values in cutoff_frequencies are not sorted by increasing
            frequencies.

        Returns
        -------
        None. Filtered signal in the 'waveform' attribute of the Sound object.
        """
        if self._filter_applied is False:
            my_filter = Filter(filter_type, cutoff_frequencies, order)
            self._waveform = my_filter.apply(self._waveform,
                                             self._waveform_sampling_frequency)
            self._filter_applied = True
            self._filter_params = my_filter
        else:
            raise ValueError('This signal has been filtered already. Cannot'
                             + ' filter twice.')

    def plot_waveform(self, unit='sec', newfig=False, title=''):
        """
        Plot waveform of the audio signal.

        PLots the waveform of the audio signal. Both the plot title and time
        units can be asjusted. The plot can be displayed on a new or an 
        existing figure.

        Parameters
        ----------
        unit : str, optional
            Time units to use. Can be either 'sec' for seconds, or 'samp' for
            samples. The default is 'sec'.
        newfig : bool, optional
            PLots on a new figure if set to True. The default is False.
        title : str, optional
            Title of the plot. The default is ''.

        Raises
        ------
        ValueError
            If the waveform attribute is empty.

        Returns
        -------
        None.

        """
        if len(self._waveform) == 0:
            raise ValueError('Cannot plot, waveform data enpty. Use Sound.read'
                             + ' to load the waveform')
        if unit == 'sec':
            axis_t = np.arange(0, len(self._waveform)
                               / self._waveform_sampling_frequency, 1
                               / self._waveform_sampling_frequency)
            xlabel = 'Time (sec)'
        elif unit == 'samp':
            axis_t = np.arange(0, len(self._waveform), 1)
            xlabel = 'Time (sample)'
        if newfig:
            plt.figure()
        axis_t = axis_t[0:len(self._waveform)]
        plt.plot(axis_t, self._waveform, color='black')
        plt.xlabel(xlabel)
        plt.ylabel('Amplitude')
        plt.title(title)
        plt.axis([axis_t[0], axis_t[-1],
                  min(self._waveform),
                  max(self._waveform)])
        plt.grid()
        plt.show()

    def select_snippet(self, chunk):
        """
        Select section of the loaded waveform.

        Create a new Sound object from a section of the sound data laoded.

        Parameters
        ----------
        chunk : list
            List of two int values representing the [start time, stop time] of
            the sound data to select, in samples. Start time must be smaller 
            than stop time.

        Raises
        ------
        ValueError
            If chunk has only one value
            If the start time is greater tahn the stop time
            If the start or stop times fall outside of the wavform limits.

        Returns
        -------
        snippet : Sound obj
            Sound object with the selected audio data.

        """
        if len(chunk) != 2:
            raise ValueError('Chunk should be a list of with 2 values: '
                             + 'chunk=[t1, t2].')
        elif chunk[0] >= chunk[1]:
            raise ValueError('Chunk[0] should be greater than chunk[1].')
        elif (chunk[0] < 0) | (chunk[0] > self.file_duration_sample):
            raise ValueError('Invalid chunk start value. The sample value '
                             + 'chunk[0] is outside of file limit.')
        elif (chunk[1] < 0) | (chunk[1] > self.file_duration_sample):
            raise ValueError('Invalid chunk stop value. The sample value '
                             + 'chunk[1] is outside of file limit.')
        snippet = copy.deepcopy(self)
        snippet._waveform = self._waveform[chunk[0]:chunk[1]]
        snippet._waveform_stop_sample = snippet._waveform_start_sample + chunk[1]
        snippet._waveform_start_sample = snippet._waveform_start_sample + chunk[0]
        snippet._waveform_duration_sample = len(snippet._waveform)
        snippet._waveform_duration_sec = snippet._waveform_duration_sec / snippet._waveform_sampling_frequency
        return snippet

    def tighten_waveform_window(self, energy_percentage):
        """
        Adjust waveform window.

        Crops the begining and end of the waveform to only capture the most
        intense part of the signal (i.e., with most energy). The percentage of
        energy is defined by the energy_percentage parameter. The attribute
        'waveform' and all its related attricbutes are updated automatically.

        Parameters
        ----------
        energy_percentage : float
            Percentage of the energy the updated waveform should have.

        Returns
        -------
        None. Updates the 'waveform' attribute alomg with all the waveform 
        -related attributes.

        """
        chunk = ecosound.core.tools.tighten_signal_limits(self._waveform, energy_percentage)
        snip = self.select_snippet(chunk)
        self.__dict__.update(snip.__dict__)

    def __len__(self):
        """Return number of samples of the waveform."""
        return self.waveform_duration_sample

    @property
    def waveform_sampling_frequency(self):
        """Return the waveform_sampling_frequency attribute."""
        return self._waveform_sampling_frequency

    @property
    def file_sampling_frequency(self):
        """Return the file_sampling_frequency attribute."""
        return self._file_sampling_frequency

    @property
    def file_duration_sample(self):
        """Return the file_duration_sample attribute."""
        return self._file_duration_sample

    @property
    def file_duration_sec(self):
        """Return the file_duration_sec attribute."""
        return self._file_duration_sec

    @property
    def channels(self):
        """Return the channels attribute."""
        return self._channels

    @property
    def channel_selected(self):
        """Return the channel_selected attribute."""
        return self._channel_selected

    @property
    def file_dir(self):
        """Return the file_dir attribute."""
        return self._file_dir

    @property
    def file_full_path(self):
        """Return the file_full_path attribute."""
        return os.path.join(self._file_dir, self._file_name) + self._file_extension

    @property
    def file_extension(self):
        """Return the file_extension attribute."""
        return self._file_extension

    @property
    def file_name(self):
        """Return the file_name attribute."""
        return self._file_name

    @property
    def waveform(self):
        """Return the waveform attribute."""
        return self._waveform

    @property
    def waveform_start_sample(self):
        """Return the waveform_start_sample attribute."""
        return self._waveform_start_sample

    @property
    def waveform_stop_sample(self):
        """Return the waveform_stop_sample attribute."""
        return self._waveform_stop_sample

    @property
    def waveform_duration_sample(self):
        """Return the waveform_duration_sample attribute."""
        return self._waveform_duration_sample

    @property
    def waveform_duration_sec(self):
        """Return the waveform_duration_sec attribute."""
        return self._waveform_duration_sec

    @property
    def filter_parameters(self):
        """Return the filter_parameters attribute."""
        if self._filter_applied:
            out = self._filter_params
        else:
            out = None
        return out

    @property
    def filter_applied(self):
        """Return the filter_applied attribute."""
        return self._filter_applied


class Filter:
    """
    Class to define a scientific filter object

    Attributes
    ----------
    type : str
        A formatted string providing the path and filename of the sound file
    freqs : list
        List with one or 2 elements defining the cut-off frequencies in Hz of
        the selected filter
    order : int
        Order of the filter

    Methods
    -------
    apply(waveform, sampling_frequency)
        Apply filter to time vector/waveform.
    coefficients(sampling_frequency)
        Defines coeeficient of the filter.

    """

    def __init__(self, type, cutoff_frequencies, order=4):
        """
        Initialize the filter.
        
        Parameters
        ----------
        type : {'bandpass', 'lowpass', 'highpass'}
            Type of filter
        cutoff_frequencies : list of float
            Cut-off frequencies of the filter sorted in increasing order (i.e.
            [lowcut, highcut]). If the filter type is 'bandpass' then 
            cutoff_frequencies must be a list of 2 floats 
            cutoff_frequencies=[lowcut, highcut], where lowcut < highcut. 
            If the filter type is 'lowpass' or 'highpass' then cutoff_frequencies
            is a list with a single float.
        order : int, optional
            Order of the filter (default is 4)

        Raises
        ------
        ValueError

            If the filter type is not set to 'bandpass', 'lowpass', or 'highpass'
            If the cutoff_frequencies has not enough of too much values for the
            filter type selected or are not sorted by increasing frequencies.
        Returns
        -------
        None. Filter object.

        """ 
        # chech filter type
        if (type == 'bandpass') | (type == 'lowpass') | (type == 'highpass') == 0:
            raise ValueError('Wrong filter type. Must be "bandpass", "lowpass"' 
                             +', or "highpass".')
        # chech freq values
        if (type == 'bandpass'):
            if len(cutoff_frequencies) != 2:
                raise ValueError('The type "bandpass" requires two frepuency '
                                 + 'values: cutoff_frequencies=[lowcut, '
                                 + 'highcut].')
            elif cutoff_frequencies[0] > cutoff_frequencies[1]:
                raise ValueError('The lowcut value should be smaller than the '
                                 + 'highcut value: cutoff_frequencies=[lowcut,'
                                 + ' highcut].')
        elif (type == 'lowpass') | (type == 'highpass'):
            if len(cutoff_frequencies) != 1:
                raise ValueError('The type "lowpass" and "highpass" require '
                                 + 'one frepuency values cutoff_frequencies='
                                 + '[cutfreq].')
        self.type = type
        self.cutoff_frequencies = cutoff_frequencies
        self.order = order

    def apply(self, waveform, sampling_frequency):
        """
        Apply filter to time series.

        Parameters
        ----------
        waveform : numpy.ndarray
            Time series to filter.
        sampling_frequency : float
            Sampling frequency of the time series to filter, in Hz.

        Returns
        -------
        numpy.ndarray
            Filtered time series.

        """
        b, a = self.coefficients(sampling_frequency)
        return spsig.lfilter(b, a, waveform)

    def coefficients(self, sampling_frequency):
        """
        Get filter coefficients.

        Parameters
        ----------
        sampling_frequency : float
            Sampling frequency of the time series to filter, in Hz.

        Returns
        -------
        b : float
            Filter coefficient b.
        a : float
            Filter coefficient a.

        """
        nyquist = 0.5 * sampling_frequency
        if self.type == 'bandpass':
            low = self.cutoff_frequencies[0] / nyquist
            high = self.cutoff_frequencies[1] / nyquist
            b, a = spsig.butter(self.order, [low, high], btype='band')
        elif self.type == 'lowpass':
            b, a = spsig.butter(self.order,
                                self.cutoff_frequencies[0]/nyquist, 'low')
        elif self.type == 'highpass':
            b, a = spsig.butter(self.order,
                                self.cutoff_frequencies[0]/nyquist, 'high')
        return b, a



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 28 06:43:58 2020

@author: xavier.mouy
"""


import xarray as xr
import pandas as pd
import numpy as np

# # Test 1: test if xarray and netcdf can hav index with same values
# df=pd.DataFrame()
# df['index']=[1,2,3,4,4,5,6,7,8,8,9,10]
# df['value1']=np.random.rand(12)
# df['value2']=np.random.rand(12)
# d=df.set_index(['index'])
# dxr=d.to_xarray()
# dxr.to_netcdf('test1.nc')
# dxr.sel(index=4)
# print('Show the 2 values with index value of 4')
# print(float(dxr.sel(index=4)['value1'][0]))
# print(float(dxr.sel(index=4)['value1'][1]))
# print('Show the 2 values with index value of 4 after writing and reimporting the xarray to netcdf')
# dxr2 = xr.open_dataset('test1.nc')
# print(float(dxr2.sel(index=4)['value1'][0]))
# print(float(dxr2.sel(index=4)['value1'][1]))

# Test 2: test if xarray csan store matrix
df=pd.DataFrame()
matrix = [np.random.rand(40,40)]*12

df = pd.DataFrame({'index':[1,2,3,4,4,5,6,7,8,8,9,10],
                   'value1': np.random.rand(12),
                   'value2': np.random.rand(12),
                   'matrix': matrix,
                   #'matrix': [],
                   })
d=df.set_index(['index'])
print('Value panda before converting to xarray')
print(d.head())
print('Type of first matrix: ',str(type(d['matrix'].iloc[0])))
# convert to xarray and extract :
dxr=d.to_xarray()
mat = dxr.sel(index=1)['matrix'].values
print('same Value from xarray ')
print(mat)
print('type: ',type(mat))
dxr.to_netcdf('test2.nc')
#dxr2 = xr.open_dataset('test2.nc')
#dxr2.sel(index=1)['matrix'].values

# Test 3: test if 2 xarrays from annot and measurements can be combined into 1
